<?php
/**
 * Copyright © Magefan (support@magefan.com). All rights reserved.
 * Please visit Magefan.com for license details (https://magefan.com/end-user-license-agreement).
 */

namespace Magefan\GuestToCustomer\Model\Order;

use \Magento\Sales\Api\OrderRepositoryInterface;
use \Magento\Sales\Model\Order\CustomerManagement as OrderCustomerManagement;
use \Magento\Customer\Api\CustomerRepositoryInterface;
use \Magento\Customer\Api\Data\AddressInterfaceFactory;
use \Magento\Customer\Api\Data\RegionInterfaceFactory;
use \Magento\Customer\Api\AccountManagementInterface ;
use \Magento\Framework\DataObject\Copy;
use \Magefan\GuestToCustomer\Model\Config;
use \Magento\Customer\Model\AddressRegistry;
use \Magefan\GuestToCustomer\Model\Config\Source\InvalidOrderAddressAction;
use \Magento\Customer\Model\AddressFactory;
use \Magento\Framework\Registry;
use \Magento\Sales\Model\Order\OrderCustomerExtractor;

/**
 * Class CustomerManagement. Manage customer info.
 */
class CustomerManagement
{
    /**
     * @var Copy
     */
    protected $objectCopyService;

    /**
     * @var OrderRepositoryInterface
     */
    protected $orderRepository;

    /**
     * @var OrderCustomerManagement
     */
    protected $orderCustomerManagement;

    /**
     * @var CustomerRepositoryInterface
     */
    protected $customerRepository;

    /**
     * @var AddressInterfaceFactory
     */
    protected $customerAddressFactory;

    /**
     * @var RegionInterfaceFactory
     */
    protected $regionFactory;

    /**
     * @var AccountManagementInterface
     */
    protected $accountManagement;

    /**
     * @var \Magento\Downloadable\Model\ResourceModel\Link\Purchased\CollectionFactory
     */
    protected $linkPurchasedCollectionFactory;

    /**
     * @var Config
     */
    protected $config;

    /**
     * @var AddressRegistry|mixed
     */
    protected $addressRegistry;

    /**
     * @var AddressFactory
     */
    private $addressFactory;

    /**
     * @var Registry
     */
    protected $registry;

    /**
     * @var OrderCustomerExtractor
     */
    protected $customerExtractor;

    /**
     * @param Copy $objectCopyService
     * @param OrderRepositoryInterface $orderRepository
     * @param OrderCustomerManagement $orderCustomerManagement
     * @param CustomerRepositoryInterface $customerRepository
     * @param AddressInterfaceFactory $customerAddressFactory
     * @param RegionInterfaceFactory $regionFactory
     * @param AccountManagementInterface $accountManagement
     * @param array $data
     * @param $linkPurchasedCollectionFactory
     * @param Config|null $config
     * @param AddressRegistry|null $addressRegistry
     * @param AddressFactory|null $addressFactory
     * @param Registry|null $registry
     */
    public function __construct(
        Copy $objectCopyService,
        OrderRepositoryInterface $orderRepository,
        OrderCustomerManagement $orderCustomerManagement,
        CustomerRepositoryInterface $customerRepository,
        AddressInterfaceFactory $customerAddressFactory,
        RegionInterfaceFactory $regionFactory,
        AccountManagementInterface $accountManagement,
        array $data = [],
        $linkPurchasedCollectionFactory = null,
        Config $config = null,
        AddressRegistry $addressRegistry = null,
        AddressFactory $addressFactory = null,
        Registry $registry = null,
        OrderCustomerExtractor $orderCustomerExtractor = null
    ) {
        $this->objectCopyService = $objectCopyService;
        $this->orderRepository = $orderRepository;
        $this->customerRepository = $customerRepository;
        $this->orderCustomerManagement = $orderCustomerManagement;
        $this->customerAddressFactory = $customerAddressFactory;
        $this->regionFactory = $regionFactory;
        $this->accountManagement = $accountManagement;

        $objectManager = \Magento\Framework\App\ObjectManager::getInstance();
        $this->linkPurchasedCollectionFactory = $linkPurchasedCollectionFactory ?: $objectManager->get(
            \Magento\Downloadable\Model\ResourceModel\Link\Purchased\CollectionFactory::class
        );
        $this->config = $config ?: $objectManager->get(
            \Magefan\GuestToCustomer\Model\Config::class
        );

        $this->addressRegistry = $addressRegistry ?: $objectManager->get(
            \Magento\Customer\Model\AddressRegistry::class
        );

        $this->addressFactory = $addressFactory ?: $objectManager->get(
            \Magento\Customer\Model\AddressFactory::class
        );

        $this->registry = $registry ?: \Magento\Framework\App\ObjectManager::getInstance()
            ->get(Registry::class);

        $this->customerExtractor = $orderCustomerExtractor ?: \Magento\Framework\App\ObjectManager::getInstance()
            ->get(OrderCustomerExtractor::class);
    }

    /**
     * Deprecated
     * @param $orderId
     * @return int|null
     * @throws \Magento\Framework\Exception\AlreadyExistsException
     * @throws \Magento\Framework\Exception\InputException
     * @throws \Magento\Framework\Exception\LocalizedException
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     * @throws \Magento\Framework\Exception\State\InputMismatchException
     */
    public function getCustomerId($orderId)
    {
        $this->registry->register('mfguesttocustomer_registry', true, true);
        return $this->convertGuestOrderById($orderId);
    }

    /**
     * @param $orderId
     * @return int|null
     * @throws \Magento\Framework\Exception\AlreadyExistsException
     * @throws \Magento\Framework\Exception\InputException
     * @throws \Magento\Framework\Exception\LocalizedException
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     * @throws \Magento\Framework\Exception\State\InputMismatchException
     */
    public function convertGuestOrderById($orderId)
    {
        $order = $this->orderRepository->get($orderId);
        $this->registry->register('mfguesttocustomer_registry', true, true);
        return $this->convertGuestOrder($order);
    }

    /**
     * @param Magento\Sales\Api\Data\OrderInterface $order
     * @return int|null
     * @throws \Magento\Framework\Exception\AlreadyExistsException
     * @throws \Magento\Framework\Exception\InputException
     * @throws \Magento\Framework\Exception\LocalizedException
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     * @throws \Magento\Framework\Exception\State\InputMismatchException
     */
    public function convertGuestOrder(\Magento\Sales\Api\Data\OrderInterface $order)
    {
        $customerId = $order->getCustomerId();
        if (!$customerId) {
            /* Order was placed by guest */

            if (!$order->getBillingAddress()) {
                return 0;
            }

            try {
                /* getCustomerEmail should be before getBillingAddress->getEmail() for "change order customer" method works good */
                $customerEmail = $customerEmail = $order->getCustomerEmail();

                if (!$customerEmail && $order->getBillingAddress()) {
                    $customerEmail = $order->getBillingAddress()->getEmail();
                }

                $customer = $this->customerRepository->get(
                    $customerEmail,
                    $order->getStore()->getWebsiteId()
                );
            } catch (\Exception $e) {
                // customer not not exist
                $customer = false;
            }

            if (!$customer || $this->config->isAssignOrderAddressToCustomer()) {
                foreach ($order->getAddresses() as $address) {
                    /* custom fix for empty firstname or lastname */
                    if (!$address->getFirstname() && $address->getLastname()) {
                        $name = explode(' ', $address->getLastname());
                        $address->setFirstname(
                            count($name) > 1 ? $name[1] : $name[0]
                        );
                    }

                    if (!$address->getLastname() && $address->getFirstname()) {
                        $name = explode(' ', $address->getFirstname());
                        $address->setLastname(
                            count($name) > 1 ? $name[1] : $name[0]
                        );
                    }

                    if (!$address->getLastname()) {
                        $address->setLastname('Customer');
                    }
                    if (!$address->getFirstname()) {
                        $address->setFirstname('Customer');
                    }
                    /* end custom fix */

                    /* custom fix for specific characters in Firstname or Lastname */
                    foreach(['Firstname', 'Lastname' , 'Middlename'] as $fieldKey) {
                        $getMethod = 'get' . $fieldKey;
                        $addressValue = $address->$getMethod();
                        if (!$addressValue) {
                            continue;
                        }

                        /* See Magento\Customer\Model\Validator\Name::PATTERN_NAME */
                        $disallowedCharacters = preg_replace('/(?:[\p{L}\p{M}\,\-\_\.\'\s\d]){1,255}+/u', '', $addressValue);
                        if ($disallowedCharacters) {
                            for($j = 0; $j < strlen($disallowedCharacters); $j++) {
                                $addressValue = str_replace($disallowedCharacters[$j], '', $addressValue);
                            }
                            $setMethod = 'set' . $fieldKey;
                            $address->$setMethod($addressValue);
                        }
                    }
                    /* end custom fix */

                    /* custom fix for prefixes in Firstname or Lastname *
                    $this->extractPrefixFromName($address, 'lastname');
                    $this->extractPrefixFromName($address, 'firstname');
                    /* end custom fix */
                }
            }

            if (!$customer) {
                /* Fix in case BillingAddress email is empty */
                if (!$order->getBillingAddress()->getEmail()
                    && $order->getCustomerEmail()
                ) {
                    $order->getBillingAddress()->setEmail($order->getCustomerEmail());
                }

                /* If customer with email not exist */
                /* Create new customer and add address */
                //$customer = $this->orderCustomerManagement->create($order->getId());
                $customer = $this->customerExtractor->extract($order->getId());
                $customer->setAddresses([]);
                $customer = $this->accountManagement->createAccount($customer);


                $customer = $this->customerRepository->getById($customer->getId());

                /* Use this logic for M2.0-2.1 as they do not have OrderCustomerExtractor */
                $store = $order->getStore();
                if ($store && $store->getWebsiteId()) {
                    $customer->setWebsiteId($store->getWebsiteId());
                }
                $customer->setStoreId($order->getStoreId());
                /* End logic for M2.0-2.1 */

                if ($customerGroupId = $this->config->getCustomerGroupId()) {
                    $customer->setGroupId($customerGroupId);
                }

                /* Subscribe customer to newsletter */
                if ($this->config->isNewsletterSubscriptionEnabled()) {
                    $extensionAttributes = $customer->getExtensionAttributes();
                    $extensionAttributes->setIsSubscribed(true);
                    $customer->setExtensionAttributes($extensionAttributes);
                }
            }

            if ($this->config->isAssignOrderAddressToCustomer()) {

                /* Add address to the customer account */
                $customerAddresses = (array)$customer->getAddresses();
                foreach ($customerAddresses as $address) {
                    $addressModel = $this->addressRegistry->retrieve($address->getId());
                    $addressModel->setShouldIgnoreValidation(true);
                }

                $orderAddresses = [];
                $shippingAddress = $order->getShippingAddress();
                $billingAddress = $order->getBillingAddress();

                if ($shippingAddress) {
                    $orderAddresses['shipping'] = $shippingAddress;
                    if (!$this->isAddressesEqual($shippingAddress, $billingAddress)) {
                        $orderAddresses['billing'] = $billingAddress;
                    }
                } else {
                    $orderAddresses['billing'] = $billingAddress;
                }

                /* Fix One or more input exceptions have occurred. */
                if ($shippingAddress) {
                    $addressFields = ['street', 'region', 'city', 'postcode'];
                    foreach ($addressFields as $addressField) {
                        if (!$billingAddress->getData($addressField)) {
                            $getMethod = 'get' . ucfirst($addressField);
                            $setMethod = 'set' . ucfirst($addressField);
                            $sValue = $shippingAddress->$getMethod();
                            if ($sValue) {
                                $billingAddress->$setMethod($sValue);
                            }
                        }
                    }
                }
                /* End fix */

                foreach ($orderAddresses as $addressType => $orderAddress) {
                    foreach ($customerAddresses as $customerAddress) {
                        if ($this->isAddressesEqual($orderAddress, $customerAddress)) {
                            //skip duplicates
                            continue 2;
                        }
                    }

                    /* Fix invalid street lines count */
                    $street = $orderAddress->getStreet();
                    if ($street && is_array($street)) {
                        $maxStreetLines = $this->config->getConfig('customer/address/street_lines', $order->getStoreId());

                        $streetLines = count($street);
                        if ($streetLines > $maxStreetLines) {
                            for ($i = $maxStreetLines; $i < $streetLines; $i++) {
                                $street[$maxStreetLines - 1] = ' ' . $street[$i];
                                unset($street[$i]);
                            }
                            $orderAddress->setStreet($street);
                        }
                    }

                    /* Create new address */
                    $newAddressData = $this->objectCopyService->copyFieldsetToTarget(
                        'order_address',
                        'to_customer_address',
                        $orderAddress,
                        []
                    );
                    $newAddressData = $this->validateAddressCustomAttributes($newAddressData);
                    $newCustomerAddress = $this->customerAddressFactory->create(['data' => $newAddressData]);
                    if (is_string($orderAddress->getRegion())) {
                        /** @var \Magento\Customer\Api\Data\RegionInterface $region */
                        $region = $this->regionFactory->create();
                        $region->setRegion($orderAddress->getRegion());
                        $region->setRegionCode($orderAddress->getRegionCode());
                        $region->setRegionId($orderAddress->getRegionId());
                        $newCustomerAddress->setRegion($region);
                    }
                    /* end */

                    if ('shipping' == $addressType || 1 == count($orderAddresses)) {
                        if (!$customer->getDefaultShipping()) {
                            $newCustomerAddress->setIsDefaultShipping(true);
                        }
                    }
                    if ('billing' == $addressType || 1 == count($orderAddresses)) {
                        if (!$customer->getDefaultBilling()) {
                            $newCustomerAddress->setIsDefaultBilling(true);
                        }
                    }

                    if (InvalidOrderAddressAction::SKIP == $this->config->invalidOrderAddressAction()) {

                        /** @var \Magento\Customer\Model\Address $addressModel */
                        $newCustomerAddressModel = $this->addressFactory->create();
                        $newCustomerAddressModel->updateData($newCustomerAddress);

                        if (true !== $newCustomerAddressModel->validate()) {
                            continue;
                        }
                    }

                    /* Add new address */
                    $customerAddresses[] = $newCustomerAddress;
                }

                /* Save customer address */
                $customer->setAddresses($customerAddresses);
            }

            try {
                $this->customerRepository->save($customer);
            } catch (\Exception $e) {
                $order
                    ->setCustomerId(null)
                    ->setCustomerIsGuest(1)
                    ->addCommentToStatusHistory(sprintf(__("Guest order assigned to customer: %s"), $customer->getEmail()));
                $this->orderRepository->save($order);
                throw $e;
            }

            $customerId = $customer->getId();

            /* Update order data */
            $order
                ->setCustomerId($customerId)
                ->setCustomerIsGuest(0)
                ->setCustomerFirstname($customer->getFirstname())
                ->setCustomerLastname($customer->getLastname())
                ->addCommentToStatusHistory(sprintf(__("Guest order assigned to customer: %s"), $customer->getEmail()));

            if ($this->config->ifChangeOrderCustomerGroup()) {
                $order->setCustomerGroupId($customer->getGroupId());
            }

            $this->orderRepository->save($order);

            /* Downloadable Link Purchased */
            $this->assignPurchasedLinks($order->getId(), $customer->getId());

            /* Fix for Stripe Payment Method */
            $this->assignStripeIntegrationPayments($order, $customer);
        }

        return $customerId;
    }

    /**
     * @param \Magento\Sales\Api\Data\OrderInterface $order
     * @param \Magento\Customer\Api\Data\CustomerInterface $newCustomer
     * @throws \Magento\Framework\Exception\AlreadyExistsException
     * @throws \Magento\Framework\Exception\InputException
     * @throws \Magento\Framework\Exception\LocalizedException
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     * @throws \Magento\Framework\Exception\State\InputMismatchException
     */
    public function changeOrderCustomer(
        \Magento\Sales\Api\Data\OrderInterface $order,
        \Magento\Customer\Api\Data\CustomerInterface $newCustomer
    ) {
        $order->setCustomerFirstname($newCustomer->getFirstname());
        $order->setCustomerLastname($newCustomer->getLastname());
        $order->setCustomerMiddlename($newCustomer->getMiddlename());
        $order->setCustomerPrefix($newCustomer->getPrefix());
        $order->setCustomerSuffix($newCustomer->getSuffix());
        $order->setCustomerEmail($newCustomer->getEmail());

        if ($oldCustomerId = $order->getCustomerId()) {
            $order->setCustomerId($newCustomer->getId());
            $this->orderRepository->save($order);
            $this->assignPurchasedLinks($order->getId(), $newCustomer->getId(), $oldCustomerId);
        } else {
            $this->registry->register('mfguesttocustomer_registry', true, true);
            $this->convertGuestOrder($order);
        }
    }

    /**
     * @param $orderId
     * @param $customerId
     * @param null $oldCustomerId
     */
    public function assignPurchasedLinks($orderId, $customerId, $oldCustomerId = null)
    {
        try {
            $linkPurchasedCollection = $this->linkPurchasedCollectionFactory->create()
                ->addFieldToFilter('order_id', ['eq' => $orderId]);
            if (null === $oldCustomerId) {
                $linkPurchasedCollection->addFieldToFilter('customer_id', ['null' => true]);
            } else {
                $linkPurchasedCollection->addFieldToFilter('customer_id', $oldCustomerId);
            }

            foreach ($linkPurchasedCollection as $link) {
                $link->setData('customer_id', $customerId);
                $link->getResource()->save($link);
            }
        } catch (\Exception $e) {
            // for case when donwloadble module is not enabled
        }
    }

    /**
     * @param $order
     * @param $customer
     */
    public function assignStripeIntegrationPayments($order, $customer)
    {
        try {
            $payment = $order->getPayment();
            if (!$payment) {
                return;
            }

            if ($payment->getMethod() != 'stripe_payments') {
                return;
            }

            $objectManager = \Magento\Framework\App\ObjectManager::getInstance();

            $tokenHelper = class_exists(\StripeIntegration\Payments\Helper\Token::class)
                ? $objectManager->get(\StripeIntegration\Payments\Helper\Token::class)
                : $objectManager->get(\StripeIntegration\Payments\Helper\Generic::class);

            $token = $tokenHelper->cleanToken($payment->getLastTransId());

            $charge = $objectManager->get(\StripeIntegration\Payments\Helper\Api::class)
                ->retrieveCharge($token);

            $stripeId = $charge->customer;

            if ($stripeId) {
                $stripeCustomer = $objectManager->create(\StripeIntegration\Payments\Model\StripeCustomer::class)
                    ->load($stripeId, 'stripe_id');

                if ($stripeCustomer->getId() && !$stripeCustomer->getCustomerId()) {
                    $stripeCustomer->setCustomerId($customer->getId())->save();
                }
            }
        } catch (\Exception $e) {
            // for case when stripe_payments module is not enabled
        }
    }

    /**
     * Return true if 2 addresses are the same
     * @param $address1
     * @param $address2
     * @return bool
     */
    protected function isAddressesEqual($address1, $address2)
    {
        $methods = ['getRegionId', 'getCountryId', 'getStreet', 'getCompany', 'getTelephone', 'getPostcode',
            'getCity', 'getFirstname', 'getLastname', 'getMiddlename', 'getPrefix', 'getSuffix', 'getVatId'];

        foreach ($methods as $method) {
            $v1 = $address1->{$method}();
            $v2 = $address2->{$method}();
            if ('0' == $v1) {
                $v1 = 0;
            }
            if ('0' == $v2) {
                $v2 = 0;
            }
            if ($v1 != $v2) {
                return false;
            }
        }

        return true;
    }

    /**
     * Validate address custom attributes
     * @param $addressAttributes
     * @return array
     */
    private function validateAddressCustomAttributes($addressAttributes)
    {
        if (isset($addressAttributes['custom_attributes']) && is_array($addressAttributes['custom_attributes'])) {
            $customAttributes = $addressAttributes['custom_attributes'];
            foreach ($customAttributes as $key => $value) {
                if (null === $value || is_object($value)) {
                    unset($addressAttributes['custom_attributes'][$key]);
                }
            }
        }
        return $addressAttributes;
    }

    /**
     * @param $address
     * @param $type
     *
    private function extractPrefixFromName($address, $type)
    {
        $name = '';
        $namePrefix = '';
        if ($type === 'lastname') {
            $name = strtolower($address->getLastname());
        } elseif ($type === 'firstname') {
            $name = strtolower($address->getFirstname());
        }

        $namePrefixesList = ["a", "ab", "af", "ap", "abu", "aït", "al", "alam", "at", "ath", "aust", "austre", "bar",
            "bath", "bat", "ben", "bin", "ibn", "bet", "bint", "binti", "binte", "da", "das", "degli", "del", "dele",
            "della", "di", "dos", "du", "e", "el", "fetch", "vetch", "fitz", "i", "ka", "kil", "gil", "la", "le",
            "lille", "lu", "m'", "mac", "mc", "mck", "mhic", "mic", "mala", "mellom", "myljom", "na", "ned", "nedre",
            "neder", "ngā", "nic", "ni", "ní", "nin", "nord", "norr", "nordre", "ny", "o", "ó", "ua", "ui", "uí", "opp",
            "upp", "öfver", "ofver", "ost", "öst", "öster", "oster", "øst", "østre", "ostre", "över", "over", "øvste",
            "ovste", "øvre", "ovre", "øver", "öz", "oz", "pour", "putra", "putera", "putri", "puteri", "'s", "setia",
            "setya", "stor", "söber", "sober", "sor", "sør", "sønder", "sonder", "syr", "søndre", "sondre", "syndre",
            "søre", "sore", "'t", "te", "ter", "tre", "den", "het", "'t", "de", "der", "van", "väst", "vast", "väster ",
            "vaster", "verch", "erch", "vest", "vestre", "vesle", "vetle", "van", "war", "zu"];

        $nameParts = explode(' ', $name);

        foreach ($nameParts as $key => $namePart) {
            if (in_array($namePart, $namePrefixesList)) {
                if (!$address->getPrefix()) {
                    $namePrefix = $namePart;
                } else {
                    $namePrefix .= ' ' . $namePart;
                }
            }
        }
        $name = str_replace($namePrefix, '', $name);
        $name = trim($name);
        if ($type === 'lastname') {
            $address->setLastname(ucfirst($name));
        } elseif ($type === 'firstname') {
            $address->setFirstname(ucfirst($name));
        }
        $address->setPrefix($namePrefix);
    }
     */
}
