<?php
/**
 * Copyright © Magefan (support@magefan.com). All rights reserved.
 * Please visit Magefan.com for license details (https://magefan.com/end-user-license-agreement).
 */

namespace Magefan\GuestToCustomer\Model\Order;

use Magefan\GuestToCustomer\Model\Order\CustomerManagement;
use Magento\Framework\App\Area;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Sales\Model\ResourceModel\Order\CollectionFactory as OrderCollectionFactory;
use Magento\Customer\Model\CustomerRegistry;
use Magento\Framework\Mail\Template\TransportBuilder;
use Magento\Customer\Api\CustomerRepositoryInterface;
use Magento\Framework\Intl\DateTimeFactory;
use Magento\Framework\Stdlib\DateTime;
use Magento\Framework\Math\Random;
use Magento\Sales\Model\ResourceModel\Order\Collection as OrderCollection;
use Magento\Sales\Api\OrderRepositoryInterface;
use Magento\Store\Model\App\Emulation;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Store\Model\ScopeInterface;
use Magento\Framework\Registry;
use Psr\Log\LoggerInterface;

class GuestOrderManagement
{
    const XML_PATH_ASSIGN_GUEST_ORDERS_TO_CUSTOMER_EMAIL_TEMPLATE = 'mfguesttocustomer/general/assign_guest_orders_to_customer_account_confirmation_email';

    /**
     * @var OrderCollectionFactory
     */
    private $orderCollectionFactory;

    /**
     * @var CustomerRegistry
     */
    private $customerRegistry;

    /**
     * @var TransportBuilder
     */
    private $transportBuilder;

    /**
     * @var CustomerRepositoryInterface
     */
    private $customerRepository;

    /**
     * @var DateTimeFactory
     */
    private $dateTimeFactory;

    /**
     * @var Random
     */
    private $mathRandom;

    /**
     * @var OrderRepositoryInterface
     */
    private $orderRepository;

    /**
     * @var CustomerManagement
     */
    private $customerManagement;

    /**
     * @var Emulation
     */
    private $emulation;

    /**
     * @var ScopeConfigInterface
     */
    private $scopeConfig;

    /**
     * @var Registry
     */
    protected $registry;

    /**
     * @var LoggerInterface
     */
    protected $logger;

    /**
     * GuestOrderManagement constructor.
     * @param OrderCollectionFactory $orderCollectionFactory
     * @param CustomerRegistry $customerRegistry
     * @param TransportBuilder $transportBuilder
     * @param CustomerRepositoryInterface $customerRepository
     * @param DateTimeFactory $dateTimeFactory
     * @param Random $mathRandom
     * @param OrderRepositoryInterface $orderRepository
     * @param \Magefan\GuestToCustomer\Model\Order\CustomerManagement $customerManagement
     * @param Emulation $emulation
     * @param ScopeConfigInterface $scopeConfig
     * @param Registry|null $registry
     * @param LoggerInterface|null $logger
     */
    public function __construct(
        OrderCollectionFactory $orderCollectionFactory,
        CustomerRegistry $customerRegistry,
        TransportBuilder $transportBuilder,
        CustomerRepositoryInterface $customerRepository,
        DateTimeFactory $dateTimeFactory,
        Random $mathRandom,
        OrderRepositoryInterface $orderRepository,
        CustomerManagement $customerManagement,
        Emulation $emulation,
        ScopeConfigInterface $scopeConfig,
        Registry $registry = null,
        LoggerInterface $logger = null
    ) {
        $this->orderCollectionFactory = $orderCollectionFactory;
        $this->customerRegistry = $customerRegistry;
        $this->transportBuilder = $transportBuilder;
        $this->customerRepository = $customerRepository;
        $this->dateTimeFactory = $dateTimeFactory;
        $this->mathRandom = $mathRandom;
        $this->orderRepository = $orderRepository;
        $this->customerManagement = $customerManagement;
        $this->emulation = $emulation;
        $this->scopeConfig = $scopeConfig;
        $this->registry = $registry ?: \Magento\Framework\App\ObjectManager::getInstance()
            ->get(Registry::class);
        $this->logger = $logger ?: \Magento\Framework\App\ObjectManager::getInstance()
            ->get(LoggerInterface::class);
    }

    /**
     * Return true if guest order(s) has been assigned to customer account
     * @param int $customerId
     * @param string $token
     * @return bool
     * @throws \Magento\Framework\Exception\LocalizedException
     * @throws NoSuchEntityException
     */
    public function assignGuestOrdersToCustomerAccount(int $customerId, string $token): bool
    {
        $result = false;

        if ($this->isActionValid($customerId, $token)) {

            $customer = $this->customerRepository->getById($customerId);

            try {
                if ($collection = $this->getGuestOrdersByEmail($customer->getEmail())) {
                    $this->registry->register('mfguesttocustomer_registry', true, true);
                    foreach ($collection as $order) {
                        $this->customerManagement->convertGuestOrder($order);
                    }

                    $result = true;
                }

            } catch (\Exception $e) {
                $this->logger->critical($e->getMessage());
            }
        }

        return $result;
    }

    /**
     * @param string $customerEmail
     * @return OrderCollection
     */
    private function getGuestOrdersByEmail(string $customerEmail): OrderCollection
    {
        return $this->orderCollectionFactory->create()
            ->addFieldToSelect('*')
            ->addAttributeToFilter('customer_is_guest', 1)
            ->addAttributeToFilter('customer_email', $customerEmail);
    }

    /**
     * Return true if exists orders with the same email as in customer that was placed under the guest account
     * @param $customerEmail
     * @return bool
     */
    public function isCustomerHasUnassignedGuestOrders(string $customerEmail): bool
    {
        return (bool) $this->getGuestOrdersByEmail($customerEmail)->getSize();
    }

    /**
     * @param int $customerId
     * @throws \Magento\Framework\Exception\InputException
     * @throws \Magento\Framework\Exception\LocalizedException
     * @throws \Magento\Framework\Exception\MailException
     * @throws NoSuchEntityException
     * @throws \Magento\Framework\Exception\State\InputMismatchException
     */
    public function sendAssignLink(int $customerId)
    {
        $customer = $this->customerRepository->getById($customerId);
        $customerSecure = $this->customerRegistry->retrieveSecureData($customerId);
        $rp_token = $customerSecure->getRpToken();

        if (!$rp_token) {
            $rp_token = $this->mathRandom->getUniqueHash();
            $customerSecure->setRpToken($rp_token);
            $customerSecure->setRpTokenCreatedAt($this->dateTimeFactory->create()->format(DateTime::DATETIME_PHP_FORMAT));
            $this->customerRepository->save($customer);
        }

        $templateParams = [
            'customer' => [
                'first_name' => $customer->getFirstName(),
                'id' => $customer->getId(),
                'rp_token' => $rp_token,
            ]
        ];

        $this->sendEmailTemplate(
            $customer,
            $this->scopeConfig->getValue(self::XML_PATH_ASSIGN_GUEST_ORDERS_TO_CUSTOMER_EMAIL_TEMPLATE),
            $templateParams,
            'sales'
        );
    }

    /**
     * @param int $customerId
     * @param string $token
     * @return bool
     * @throws NoSuchEntityException
     */
    private function isActionValid(int $customerId, string $token) : bool
    {
        return ($this->getToken($customerId) == $token);
    }

    /**
     * @param $customer
     * @param string $template
     * @param array $templateParams
     * @param string $scope
     * @throws \Magento\Framework\Exception\LocalizedException
     * @throws \Magento\Framework\Exception\MailException
     */
    protected function sendEmailTemplate($customer, string $template, array $templateParams = [], string $scope = 'general')
    {
        $this->emulation->startEnvironmentEmulation($customer->getStoreId(), 'frontend');

        $sender = [
            'name' => $this->getConfigValue('trans_email/ident_general/name', $customer->getStoreId()),
            'email' => $this->getConfigValue('trans_email/ident_general/email', $customer->getStoreId())
        ];


        $transport = $this->transportBuilder->setTemplateIdentifier($template)
            ->setTemplateOptions(
                [
                    'area' => Area::AREA_FRONTEND,
                    'store' => $customer->getStoreId()
                ]
            )
            ->setTemplateVars($templateParams)
            //->setFromByScope($scope) //is not available in some Magento versions
            ->addTo($customer->getEmail())
            ->setFrom($sender)
            ->getTransport();

        $transport->sendMessage();

        $this->emulation->stopEnvironmentEmulation();
    }

    /**
     * @param $path
     * @return mixed
     */
    private function getConfigValue($path, $storeId)
    {
        return $this->scopeConfig->getValue($path, ScopeInterface::SCOPE_STORE, $storeId);
    }

    /**
     * @param int $customerId
     * @return string
     * @throws NoSuchEntityException
     */
    public function getToken(int $customerId)
    {
        return $this->customerRegistry->retrieveSecureData($customerId)->getRpToken();
    }
}
