<?php
/**
 * Copyright © Magefan (support@magefan.com). All rights reserved.
 * Please visit Magefan.com for license details (https://magefan.com/end-user-license-agreement).
 */

namespace Magefan\GuestToCustomer\Observer;

use Magento\Framework\Event\ObserverInterface;
use Magento\Framework\Event\Observer;
use Psr\Log\LoggerInterface;
use Magefan\GuestToCustomer\Model\Config;
use Magefan\GuestToCustomer\Model\Order\CustomerManagement;
use Magento\Framework\Indexer\IndexerRegistry;
use Magento\Framework\Registry;
use Magento\Framework\App\ProductMetadataInterface;
use Magento\Sales\Model\OrderFactory;

/**
 * Class SalesOrderSaveAfter save order with new customer.
 */
class SalesOrderSaveAfter implements ObserverInterface
{
    /**
     * @var Config
     */
    protected $config;

    /**
     * @var LoggerInterface
     */
    protected $logger;

    /**
     * @var IndexerRegistry
     */
    protected $indexerRegistry;

    /**
     * @var Registry
     */
    protected $registry;

    /**
     * @var CustomerManagement
     */
    protected $customerManagement;

    /**
     * @var ProductMetadataInterface
     */
    protected $productMetadata;

    /**
     * @var OrderFactory
     */
    protected $orderFactory;

    /**
     * @param Config $config
     * @param LoggerInterface $logger
     * @param CustomerManagement $customerManagement
     * @param IndexerRegistry|null $indexerRegistry
     * @param Registry|null $registry
     * @param ProductMetadataInterface|null $productMetadata
     * @param OrderFactory|null $orderFactory
     */
    public function __construct(
        Config $config,
        LoggerInterface $logger,
        CustomerManagement $customerManagement,
        IndexerRegistry $indexerRegistry = null,
        Registry $registry = null,
        ProductMetadataInterface $productMetadata = null,
        OrderFactory $orderFactory = null
    ) {
        $this->config = $config;
        $this->logger = $logger;
        $this->customerManagement = $customerManagement;
        $this->indexerRegistry = $indexerRegistry ?: \Magento\Framework\App\ObjectManager::getInstance()
            ->get(\Magento\Framework\Indexer\IndexerRegistry::class);
        $this->registry = $registry ?: \Magento\Framework\App\ObjectManager::getInstance()
            ->get(Registry::class);
        $this->productMetadata = $productMetadata ?: \Magento\Framework\App\ObjectManager::getInstance()
            ->get(ProductMetadataInterface::class);
        $this->orderFactory = $orderFactory ?: \Magento\Framework\App\ObjectManager::getInstance()
            ->create(OrderFactory::class);
    }

    /**
     * @param \Magento\Framework\Event\Observer $observer
     */
    public function execute(Observer $observer)
    {
        $order = $observer->getEvent()->getOrder();
        if (!$order || !$order->getId() || $order->getCustomerId()) {
            return;
        }

        try {
            $storeId = $order->getStoreId();

            if ($this->config->isEnabled($storeId)
                && $this->config->isAutoconvertEnabled($storeId)
                && array_intersect($this->config->getActomaticallyConvertGuestOnOrderStatus(), [0, $order->getStatus()])
            ) {

                if (version_compare($this->productMetadata->getVersion(),  '2.4.3', '<')) {
                    /* Fix for old Magento bug fixed in this commit
                    https://github.com/magento/magento2/commit/96dc6939b28752c7ad5d2e823082a8a1815e6bf1
                    */
                    $orderModel = $this->orderFactory->create()->load($order->getId());
                    if (!$orderModel->getPayment()) {
                        return;
                    }
                }

                /*convert guest to customer*/
                $this->registry->register('mfguesttocustomer_registry', true, true);
                $this->customerManagement->convertGuestOrderById($order->getId());
                /*reindex new customer */
                if ($order->getCustomerId()) {
                    // Get index for each action separately
                    //$this->indexerRegistry->get(\Magento\Customer\Model\Customer::CUSTOMER_GRID_INDEXER_ID)->invalidate();
                    $this->indexerRegistry->get(\Magento\Customer\Model\Customer::CUSTOMER_GRID_INDEXER_ID)->reindexRow($order->getCustomerId());
                }
            }
        } catch (\Exception $e) {
            $this->logger->critical($e->getMessage());
        }
    }
}
