<?php

namespace MagePsycho\MipPricing\Model\Import;

use Magento\Framework\DB\Sql\Expression;
use Magento\Framework\Event\ManagerInterface;
use Magento\Framework\Stdlib\DateTime;
use Magento\Framework\Stdlib\DateTime\DateTime as DateTimeDateTime;
use MagePsycho\MassImporterPro\Model\Import\ErrorProcessing\ProcessingError;

use MagePsycho\MipPricing\Model\Import\Pricing\Validator\TierPrice as TierPriceValidator;
use MagePsycho\MipPricing\Model\Import\Pricing\RowValidatorInterface as ValidatorInterface;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\DB\Adapter\AdapterInterface;
use Magento\Framework\Json\Helper\Data as JsonHelper;

use MagePsycho\MassImporterPro\Helper\Data as ImportHelper;
use MagePsycho\MassImporterPro\Model\Import;
use MagePsycho\MassImporterPro\Model\Import\Entity\AbstractEntity;
use MagePsycho\MassImporterPro\Model\Import\ErrorProcessing\ProcessingErrorAggregatorInterface;
use MagePsycho\MassImporterPro\Model\ResourceModel\Helper;
use MagePsycho\MassImporterPro\Model\ResourceModel\Import\Data;
use Magento\Eav\Model\Entity\Attribute\ScopedAttributeInterface;
use Magento\Catalog\Api\Data\ProductInterface;

/**
 * @category   MagePsycho
 * @package    MagePsycho_MipPricing
 * @author     Raj KB <magepsycho@gmail.com>
 * @website    https://www.magepsycho.com
 * @license    http://opensource.org/licenses/osl-3.0.php  Open Software License (OSL 3.0)
 */
class Pricing extends AbstractEntity
{
    const ENTITY_CODE           = 'mip_pricing';

    const TABLE_MAIN            = 'catalog_product_entity';
    const TABLE_TIER_PRICE      = 'catalog_product_entity_tier_price';
    const TABLE_DECIMAL_PRICE   = 'catalog_product_entity_decimal';
    const TABLE_DATETIME        = 'catalog_product_entity_datetime';

    const COL_SKU               = 'sku';
    const COL_NAME              = 'name';
    const COL_WEBSITE_ID        = 'website_id';
    const COL_COST              = 'cost';
    const COL_PRICE             = 'price';
    const COL_MSRP              = 'msrp';
    const COL_SPECIAL_PRICE     = 'special_price';
    const COL_SPECIAL_FROM_DATE = 'special_from_date';
    const COL_SPECIAL_TO_DATE   = 'special_to_date';
    const COL_TIER_PRICE_ALL    = 'tier_price:ALL GROUPS';
    const COL_GROUP_PRICE_ALL   = 'group_price:ALL GROUPS';

    const GROUP_ALL          = 'ALL GROUPS';
    const TIER_PRICE_PREFIX  = 'tier_price';
    const GROUP_PRICE_PREFIX = 'group_price';

    const VALIDATOR_TIER_PRICE = 'validator_tier_price';

    /**
     * If we should check column names
     */
    protected $needColumnCheck = true;

    /**
     * Need to log in import history
     */
    protected $logInHistory = true;

    /**
     * Permanent entity columns.
     */
    protected $_permanentAttributes = [
        self::COL_SKU,
        self::COL_WEBSITE_ID
    ];

    protected $_specialAttributes = [
        self::COL_NAME
    ];

    /**
     * Valid column names
     */
    protected $validColumnNames = [
        self::COL_SKU,
        self::COL_WEBSITE_ID,
        self::COL_COST,
        self::COL_PRICE,
        self::COL_MSRP,
        self::COL_SPECIAL_PRICE,
        self::COL_SPECIAL_FROM_DATE,
        self::COL_SPECIAL_TO_DATE,
        self::COL_TIER_PRICE_ALL,
        self::COL_GROUP_PRICE_ALL
        // other dynamic groups are initiated by constructor
    ];

    /**
     * Attributes codes which shows as date
     *
     * @var array
     */
    protected $dateAttrCodes = [
        self::COL_SPECIAL_FROM_DATE,
        self::COL_SPECIAL_TO_DATE,
    ];

    protected $decimalAttrCodes = [
        self::COL_COST,
        self::COL_PRICE,
        self::COL_MSRP,
        self::COL_SPECIAL_PRICE
    ];

    /**
     * Validation failure message template definitions.
     *
     * @var array
     */
    protected $_messageTemplates = [
        ValidatorInterface::ERROR_SKU_IS_EMPTY          => 'SKU is empty',
        ValidatorInterface::ERROR_SKU_NOT_FOUND         => 'Product with specified SKU not found',
        ValidatorInterface::ERROR_WEBSITE_IS_EMPTY      => 'Website is empty',
        ValidatorInterface::ERROR_WEBSITE_NOT_FOUND     => 'Website doesn\'t exist',
        ValidatorInterface::ERROR_INVALID_WEBSITE       => 'Website doesn\'t exist',
        ValidatorInterface::ERROR_TIER_DATA_IS_EMPTY    => 'Tier Price data is empty',
        ValidatorInterface::ERROR_TIER_DATA_INVALID     => 'Tier Price data is invalid',
        ValidatorInterface::ERROR_GROUP_DATA_IS_EMPTY   => 'Group Price data is empty',
        ValidatorInterface::ERROR_GROUP_DATA_INVALID    => 'Group Price data is invalid',
        ValidatorInterface::ERROR_INVALID_DATE          => 'Date format is invalid',
        ValidatorInterface::ERROR_INVALID_PRICE_FORMAT  => 'Price format is invalid',
        ValidatorInterface::ERROR_INVALID_PRICE_VALUE   => 'Price value is invalid',
        ValidatorInterface::ERROR_INVALID_TO_DATE       => 'To Date should be later than or same as From Date'
    ];

    /**
     * Existing products SKU-related information in form of array:
     *
     * [SKU] => array(
     *     'type_id'        => (string) product type
     *     'attr_set_id'    => (int) product attribute set ID
     *     'entity_id'      => (int) product ID
     *     'supported_type' => (boolean) is product type supported by current version of import module
     * )
     *
     * @var array
     */
    protected $_oldSku = [];

    /**
     * Existing products SKU related info in form of array:
     * [SKU] => array(
     *      <website_id1> => (float) price,
     *      <website_id2> => (float) price,
     * )
     *
     * @var array
     */
    protected $_basePrices = [];

    protected $websitesCache = [];

    /**
     * Attribute cache
     *
     * @var array
     */
    protected $_attributeCache = [];

    /**
     * @var array
     */
    protected $_cachedSkuToDelete;

    /**
     * @var array
     */
    protected $_validators = [];

    /**
     * @var AdapterInterface
     */
    protected $_connection;

    /**
     * @var Proxy\Product\ResourceModelFactory
     */
    protected $_resourceFactory;

    /**
     * @var Proxy\Product\ResourceModel
     */
    protected $_productResource;

    /**
     * @var DateTime
     */
    protected $dateTime;

    /**
     * @var DateTimeDateTime
     */
    private $dateTimeDateTime;

    /**
     * @var \Magento\Catalog\Helper\Data
     */
    protected $_catalogData;

    /**
     * @var StoreResolver
     */
    protected $_storeResolver;

    /**
     * @var Pricing\SkuProcessor
     */
    private $skuProcessor;

    /**
     * @var Pricing\DecimalProcessor
     */
    private $decimalProcessor;

    /**
     * @var Pricing\Validator
     */
    private $validator;

    /**
     * @var ManagerInterface
     */
    private $eventManager;

    /**
     * Product entity link field
     *
     * @var string
     */
    private $productEntityLinkField;

    /**
     * Catalog product entity
     *
     * @var string
     */
    private $productEntityTableName;

    /**
     * @var string
     */
    private $productEntityDatetimeTableName;

    /**
     * @var string
     */
    private $productEntityDecimalTableName;

    /**
     * @var string
     */
    private $productEntityTierPriceTableName;

    public function __construct(
        JsonHelper $jsonHelper,
        ImportHelper $importExportData,
        Data $importData,
        ResourceConnection $resource,
        Helper $resourceHelper,
        ProcessingErrorAggregatorInterface $errorAggregator,
        DateTime $dateTime,
        DateTimeDateTime $dateTimeDateTime,
        Proxy\Product\ResourceModelFactory $resourceFactory,
        \Magento\Catalog\Helper\Data $catalogData,
        Pricing\StoreResolver $storeResolver,
        Pricing\SkuProcessor $skuProcessor,
        Pricing\DecimalProcessor $decimalProcessor,
        Pricing\Validator $validator,
        Pricing\Validator\TierPrice $tierPriceValidator,
        ManagerInterface $eventManager
    ) {
        $this->jsonHelper = $jsonHelper;
        $this->_importExportData = $importExportData;
        $this->_resourceHelper = $resourceHelper;
        $this->_dataSourceModel = $importData;
        $this->_connection = $resource->getConnection();
        $this->_resourceFactory = $resourceFactory;
        $this->errorAggregator = $errorAggregator;
        $this->dateTime = $dateTime;
        $this->dateTimeDateTime = $dateTimeDateTime;
        $this->_catalogData = $catalogData;
        $this->_storeResolver = $storeResolver;
        $this->skuProcessor = $skuProcessor;
        $this->decimalProcessor = $decimalProcessor;
        $this->eventManager = $eventManager;
        $this->validator = $validator;

        $this->validator->init($this);
        $this->_validators[self::VALIDATOR_TIER_PRICE] = $tierPriceValidator;

        $this->initMessageTemplates();
        $this->initValidColumnNamesForTierPrice();
        $this->initSkus();
        $this->initBasePrices();
    }

    /**
     * Initialize existent product SKUs.
     *
     * @return $this
     */
    protected function initSkus()
    {
        $this->_oldSku = $this->skuProcessor->reloadOldSkus()->getOldSkus();
        return $this;
    }

    protected function initBasePrices()
    {
        $this->_basePrices = $this->decimalProcessor->getBasePrices();
    }

    /**
     * Init Error Messages
     */
    private function initMessageTemplates()
    {
        foreach (array_merge($this->errorMessageTemplates, $this->_messageTemplates) as $errorCode => $message) {
            $this->getErrorAggregator()->addErrorMessageTemplate($errorCode, $message);
        }
    }

    private function initValidColumnNamesForTierPrice()
    {
        $customerGroups = $this->_getValidator(self::VALIDATOR_TIER_PRICE)->getCustomerGroups();
        foreach ($customerGroups as $groupId => $groupName) {
            $this->validColumnNames = array_merge(
                $this->validColumnNames,
                [self::TIER_PRICE_PREFIX . ':' . $groupName, self::GROUP_PRICE_PREFIX . ':' . $groupName]
            );
        }
    }

    /**
     * Entity type code getter.
     *
     * @return string
     */
    public function getEntityTypeCode()
    {
        return static::ENTITY_CODE;
    }

    /**
     * Get available columns
     *
     * @return array
     */
    public function getValidColumnNames()
    {
        return $this->validColumnNames;
    }

    public function validateRow(array $rowData, $rowNum)
    {
        if (isset($this->_validatedRows[$rowNum])) {
            // check that row is already validated
            return !$this->getErrorAggregator()->isRowInvalid($rowNum);
        }
        $this->_validatedRows[$rowNum] = true;

        $sku = $rowData[self::COL_SKU];
        if (!strlen($sku)) {
            $this->skipRow($rowNum, ValidatorInterface::ERROR_SKU_IS_EMPTY);
            return false;
        }
        if (!$this->isSkuExist($sku)) {
            $this->skipRow($rowNum, ValidatorInterface::ERROR_SKU_NOT_FOUND);
            return false;
        }

        // if product doesn't exist, need to throw critical error else all errors should be not critical.
        $errorLevel = $this->getValidationErrorLevel($sku);

        if (!$this->validator->isValid($rowData)) {
            foreach ($this->validator->getMessages() as $message) {
                $this->skipRow($rowNum, $message, $errorLevel);
            }
        }

        // SKU is specified, row is SCOPE_DEFAULT, new product block begins
        $this->_processedEntitiesCount++;

        return !$this->getErrorAggregator()->isRowInvalid($rowNum);
    }

    /**
     * Import data
     *
     * @return bool
     * @throws \Exception
     */
    protected function _importData()
    {
        if (Import::BEHAVIOR_DELETE == $this->getBehavior()) {
            $this->deletePricing();
        } elseif (Import::BEHAVIOR_REPLACE == $this->getBehavior()) {
            $this->replacePricing();
        } elseif (Import::BEHAVIOR_APPEND == $this->getBehavior()) {
            $this->savePricing();
        }

        $this->eventManager->dispatch('mip_pricing_import_finish_before', ['adapter' => $this]);

        return true;
    }

    public function deletePricing()
    {
        $this->saveAndReplacePrices();
        return $this;
    }

    public function savePricing()
    {
        $this->saveAndReplacePrices();
        return $this;
    }

    /**
     * Replace pricing
     *
     * @return $this
     */
    public function replacePricing()
    {
        $this->saveAndReplacePrices();
        return $this;
    }

    protected function saveAndReplacePrices()
    {
        $behavior = $this->getBehavior();

        $listSku = [];
        while ($bunch = $this->_dataSourceModel->getNextBunch()) {
            $tierPrices     = [];
            $decimalPrices  = [];
            $specialDates   = [];
            foreach ($bunch as $rowNum => $rowData) {
                if (!$this->validateRow($rowData, $rowNum)) {
                    $this->addRowError(ValidatorInterface::ERROR_SKU_IS_EMPTY, $rowNum);
                    continue;
                }
                if ($this->getErrorAggregator()->hasToBeTerminated()) {
                    $this->getErrorAggregator()->addRowToSkip($rowNum);
                    continue;
                }

                $rowSku = $rowData[self::COL_SKU];
                if (null === $rowSku) {
                    $this->getErrorAggregator()->addRowToSkip($rowNum);
                    continue;
                }

                if (!$this->isSkuExist($rowSku)) {
                    $rowSku = null;
                    // sign for child rows to be skipped
                    $this->getErrorAggregator()->addRowToSkip($rowNum);
                    continue;
                }

                $listSku[] = $rowSku;

                $preparedPriceAttrData = $this->preparePriceAttributesRow($rowData);

                $decimalPrices = array_replace_recursive($decimalPrices, $preparedPriceAttrData['decimal']);
                $specialDates = array_replace_recursive($specialDates, $preparedPriceAttrData['date']);
                $tierPrices = array_replace_recursive($tierPrices, $this->prepareTierPriceRow($rowData));
            }

            foreach ($bunch as $rowNum => $rowData) {
                if ($this->getErrorAggregator()->isRowInvalid($rowNum)) {
                    unset($bunch[$rowNum]);
                }
            }

            if (Import::BEHAVIOR_REPLACE == $behavior) {
                $this->saveSpecialDates($specialDates);
                $this->saveDecimalPrices($decimalPrices);
                $this->saveTierPrices($tierPrices);
            } elseif (Import::BEHAVIOR_APPEND == $behavior) {
                $this->saveSpecialDates($specialDates);
                $this->saveDecimalPrices($decimalPrices);
                $this->saveTierPrices($tierPrices);
            } elseif (Import::BEHAVIOR_DELETE == $behavior) {
                $this->deleteSpecialDates($specialDates);
                $this->deleteDecimalPrices($decimalPrices);
                $this->deleteTierPrices($tierPrices);
            }

            $this->eventManager->dispatch(
                'mip_pricing_import_bunch_save_after',
                ['adapter' => $this, 'bunch' => $bunch]
            );
        }

        if ($listSku) {
            $this->setUpdatedAt($listSku);
        }

        return $this;
    }

    protected function preparePriceAttributesRow(array $rowData)
    {
        $decimalPrices  = [];
        $specialDates   = [];

        $isPriceGlobal      = $this->_catalogData->isPriceGlobal();
        $rowSku = $rowData[self::COL_SKU];
        $websiteId = $rowData[self::COL_WEBSITE_ID];
        $canHaveMultiStoreIds = $this->_storeResolver->isSingleStoreMode() || empty($websiteId) ? false : true;
        $rowStoreIds  = [0];
        if ($canHaveMultiStoreIds) {
            $rowStoreIds = $this->_storeResolver->getWebsiteIdToStoreIds($websiteId);
        }
        $basePrice = $this->getBasePrice($rowSku, $websiteId);
        foreach ($rowData as $attrCode => $attrValue) {
            // Don't delete price values (even by intentionally)
            if ($this->getBehavior() == Import::BEHAVIOR_DELETE && $attrCode == self::COL_PRICE) {
                continue;
            }

            if (in_array($attrCode, array_merge($this->decimalAttrCodes, $this->dateAttrCodes))) {
                $attribute  = $this->retrieveAttributeByCode($attrCode);
                $attrId     = $attribute->getId();

                // preparing the store-ids
                // ScopedAttributeInterface::SCOPE_GLOBAL
                $isDecimalAttribute         = $this->isValidDecimalAttribute($attribute);
                $isSpecialDateAttribute     = $this->isValidDateAttribute($attribute);
                $storeIds                   = [0];
                if ($canHaveMultiStoreIds
                    && (($isDecimalAttribute || $isSpecialDateAttribute) && !$isPriceGlobal)
                    && (ScopedAttributeInterface::SCOPE_WEBSITE == $attribute->getIsGlobal()
                        || ScopedAttributeInterface::SCOPE_STORE == $attribute->getIsGlobal()
                    )
                ) {
                    $storeIds = $rowStoreIds;
                }

                // preparing attribute values
                if ($this->getBehavior() == Import::BEHAVIOR_DELETE) {
                    $attrValue = ValidatorInterface::DELETE_MARKER;
                } else {
                    $attrValue = preg_replace('/\s+/', '', (string) $attrValue);

                    // Skip for empty values
                    if (!strlen($attrValue)) {
                        continue;
                    }

                    // Price value marked with x or X will be operated separately
                    if (strtolower($attrValue) != ValidatorInterface::DELETE_MARKER) {
                        if ($isSpecialDateAttribute) {
                            $attrValue = $this->dateTime->formatDate($attrValue, false);
                        } elseif ($isDecimalAttribute) {
                            $attrValue = $this->decimalProcessor->isPureNumeric($attrValue)
                                ? $attrValue
                                : new Expression(
                                    $this->decimalProcessor->parseValue($attrValue, $attrCode, $basePrice)
                                );
                        }
                    }
                }

                foreach ($storeIds as $storeId) {
                    if ($isSpecialDateAttribute) {
                        $specialDates[$rowSku][$attrId][$storeId] = $attrValue;
                    }
                    if ($isDecimalAttribute) {
                        $decimalPrices[$rowSku][$attrId][$storeId] = $attrValue;
                    }
                }
            }
        }
        return [
            'decimal'   => $decimalPrices,
            'date'      => $specialDates,
        ];
    }

    protected function prepareTierPriceRow(array $rowData)
    {
        $tierPrices  = [];

        $rowSku     = $rowData[self::COL_SKU];
        $websiteId  = $rowData[self::COL_WEBSITE_ID];
        $basePrice  = $this->getBasePrice($rowSku, $websiteId);
        foreach ($rowData as $attrCode => $attrValue) {
            if (preg_match('/' . self::TIER_PRICE_PREFIX . ':(.*)/', $attrCode, $matches)
                || preg_match('/' . self::GROUP_PRICE_PREFIX . ':(.*)/', $attrCode, $matches)
            ) {
                $customerGroup   = isset($matches[1]) ? trim($matches[1]) : null;
                if (empty($customerGroup)) {
                    //@todo logging
                    continue;
                }
                $allGroups = $customerGroup == self::GROUP_ALL ? 1 : 0;

                // For delete behavior no need to parse values
                if ($this->getBehavior() == Import::BEHAVIOR_DELETE) {
                    $tierPrices[$rowSku][] = [
                        'website_id'        => $websiteId,
                        'customer_group_id' => $this->getCustomerGroupId($customerGroup),
                        'all_groups'        => $allGroups,
                        'value'             => ValidatorInterface::DELETE_MARKER
                    ];
                    continue;
                }

                // Prepare data for create/update
                $attrValue = preg_replace('/\s+/', '', (string) $attrValue);
                $attrValue = trim($attrValue, TierPriceValidator::MULTI_VALUES_SEPARATOR);
                $tierQtyToPrices = explode(TierPriceValidator::MULTI_VALUES_SEPARATOR, $attrValue);

                foreach ($tierQtyToPrices as $tierQtyToPrice) {
                    if (!strlen($tierQtyToPrice)) {
                        //@todo logging
                        continue;
                    }

                    if (preg_match('/' . self::TIER_PRICE_PREFIX . ':(.*)/', $attrCode)) {
                        $tierQtyPriceArray = explode(
                            TierPriceValidator::QTY_TO_PRICE_SEPARATOR,
                            $tierQtyToPrice
                        );
                        list($tierQty, $tierPrice) = count($tierQtyPriceArray) !== 2
                            ? [ValidatorInterface::DELETE_QTY_VALUE, $tierQtyToPrice]
                            : $tierQtyPriceArray;
                    } else {
                        // Group price is the special case of tier price with Qty = 1
                        list($tierQty, $tierPrice) = [1, $tierQtyToPrice];
                    }

                    if (empty($tierQty) || empty($tierPrice)) {
                        //@todo logging
                        continue;
                    }

                    // Price value marked with x or X will be operated separately
                    if (strtolower($tierPrice) == ValidatorInterface::DELETE_MARKER) {
                        $tierPriceValue  = $tierPrice;
                        $percentageValue = null;
                    } else {
                        $tierPriceValue = $this->decimalProcessor->isFixedType($tierPrice)
                            ? (
                                $this->decimalProcessor->isPureNumeric($tierPrice)
                                ? $tierPrice
                                : new Expression(
                                    $this->decimalProcessor->parseValue($tierPrice, self::TIER_PRICE_PREFIX, $basePrice)
                                )
                            ) : 0;
                        $percentageValue = $this->decimalProcessor->isPercentageType($tierPrice)
                            ? str_replace('%', '', $tierPrice)
                            : null;
                    }

                    $tierPrices[$rowSku][] = [
                        'all_groups'        => $allGroups,
                        'customer_group_id' => $this->getCustomerGroupId($customerGroup),
                        'qty'               => $tierQty,
                        'value'             => $tierPriceValue,
                        'percentage_value'  => $percentageValue,
                        'website_id'        => $websiteId
                    ];
                }
            }
        }
        return $tierPrices;
    }

    protected function isValidDecimalAttribute($attribute)
    {
        return ('decimal' == $attribute->getBackendType()
            && in_array($attribute->getAttributeCode(), $this->decimalAttrCodes)
        );
    }

    protected function isValidDateAttribute($attribute)
    {
        return (
            'datetime' == $attribute->getBackendType()
            && in_array($attribute->getAttributeCode(), $this->dateAttrCodes)
        );
    }

    protected function saveSpecialDates(array $data)
    {
        if (empty($data)) {
            return $this;
        }

        $tableName = $this->getProductEntityDatetimeTableName();
        $linkField = $this->getProductEntityLinkField();
        $skus = $this->getOldSku();
        $updateData = [];
        $deleteData = [];
        foreach ($data as $sku => $rows) {
            if (!isset($skus[strtolower($sku)])) {
                continue;
            }

            $productId = $skus[strtolower($sku)][$linkField];
            foreach ($rows as $attributeId => $storeValues) {
                foreach ($storeValues as $storeId => $storeValue) {
                    $operationData = [
                        $linkField      => $productId,
                        'attribute_id'  => $attributeId,
                        'store_id'      => $storeId,
                        'value'         => $storeValue,
                    ];

                    if (strtolower($storeValue) == ValidatorInterface::DELETE_MARKER) {
                        // Prepare the data compatible with deleteSpecialDates()
                        $deleteData[$sku][$attributeId][$storeId] = $storeValue;
                    } else {
                        $updateData[] = $operationData;
                    }
                }
            }
        }

        if ($deleteData) {
            $this->deleteSpecialDates($deleteData);
        }
        if ($updateData) {
            $this->countItemsUpdated += $this->_connection->insertOnDuplicate($tableName, $updateData, ['value']);
        }

        return $this;
    }

    protected function deleteSpecialDates(array $data)
    {
        if (empty($data)) {
            return $this;
        }

        $tableName = $this->getProductEntityDatetimeTableName();
        $linkField = $this->getProductEntityLinkField();
        $skus   = $this->getOldSku();
        foreach ($data as $sku => $rows) {
            if (!isset($skus[strtolower($sku)])) {
                continue;
            }

            $productId = $skus[strtolower($sku)][$linkField];
            foreach ($rows as $attributeId => $storeValues) {
                $this->countItemsDeleted += $this->_connection->delete(
                    $tableName,
                    [
                        $this->_connection->quoteInto($linkField . ' = ?', $productId),
                        $this->_connection->quoteInto('attribute_id = ?', $attributeId),
                        $this->_connection->quoteInto('store_id IN (?)', array_keys($storeValues))
                    ]
                );
            }
        }

        return $this;
    }

    protected function saveDecimalPrices(array $data)
    {
        if (empty($data)) {
            return $this;
        }

        $tableName = $this->getProductEntityDecimalTableName();
        $linkField = $this->getProductEntityLinkField();
        $skus      = $this->getOldSku();
        $updateData = [];
        $deleteData = [];
        foreach ($data as $sku => $rows) {
            if (!isset($skus[strtolower($sku)])) {
                continue;
            }

            $productId = $skus[strtolower($sku)][$linkField];
            foreach ($rows as $attributeId => $storeValues) {
                foreach ($storeValues as $storeId => $storeValue) {
                    $operationData = [
                        $linkField      => $productId,
                        'attribute_id'  => $attributeId,
                        'store_id'      => $storeId,
                        'value'         => $storeValue,
                    ];

                    if (strtolower($storeValue) == ValidatorInterface::DELETE_MARKER) {
                        // Prepare the data compatible with deleteDecimalPrices()
                        $deleteData[$sku][$attributeId][$storeId] = $storeValue;
                    } else {
                        if ($this->decimalProcessor->hasSelfUpdate($storeValue)) {
                            $this->countItemsUpdated += $this->_connection->update(
                                $tableName,
                                ['value' => $storeValue],
                                [
                                    $this->_connection->quoteInto("$linkField = ?", $productId),
                                    $this->_connection->quoteInto('attribute_id = ?', $attributeId),
                                    $this->_connection->quoteInto('store_id = ?', $storeId)
                                ]
                            );
                        } else {
                            $updateData[] = $operationData;
                        }
                    }
                }
            }
        }

        if ($deleteData) {
            $this->deleteDecimalPrices($deleteData);
        }
        if ($updateData) {
            $this->countItemsUpdated += $this->_connection->insertOnDuplicate($tableName, $updateData, ['value']);
        }

        return $this;
    }

    protected function deleteDecimalPrices(array $data)
    {
        if (empty($data)) {
            return $this;
        }

        $tableName = $this->getProductEntityDecimalTableName();
        $linkField = $this->getProductEntityLinkField();
        $skus   = $this->getOldSku();
        foreach ($data as $sku => $rows) {
            if (!isset($skus[strtolower($sku)])) {
                continue;
            }

            $productId = $skus[strtolower($sku)][$linkField];
            foreach ($rows as $attributeId => $storeValues) {
                $this->countItemsDeleted += $this->_connection->delete(
                    $tableName,
                    [
                        $this->_connection->quoteInto($linkField . ' = ?', $productId),
                        $this->_connection->quoteInto('attribute_id = ?', $attributeId),
                        $this->_connection->quoteInto('store_id IN (?)', array_keys($storeValues))
                    ]
                );
            }
        }

        return $this;
    }

    protected function saveTierPrices(array $data)
    {
        if (empty($data)) {
            return $this;
        }

        $tableName = $this->getProductEntityTierPriceTableName();
        $linkField = $this->getProductEntityLinkField();

        $updateData    = [];
        $skus       = $this->getOldSku();
        foreach ($data as $sku => $priceRows) {
            if (!isset($skus[strtolower($sku)])) {
                continue;
            }

            $productId = $skus[strtolower($sku)][$linkField];
            foreach ($priceRows as $row) {
                $row[$linkField] = $productId;

                // Individual delete based on website_id & customer_group_id
                if (Import::BEHAVIOR_APPEND != $this->getBehavior()) {
                    $this->countItemsDeleted += $this->_connection->delete(
                        $tableName,
                        [
                            $this->_connection->quoteInto($linkField . ' = ?', $productId),
                            $this->_connection->quoteInto('website_id = ?', $row['website_id']),
                            $this->_connection->quoteInto('customer_group_id = ?', $row['customer_group_id']),
                            $this->_connection->quoteInto('all_groups = ?', $row['all_groups'])
                        ]
                    );
                }

                // Individual delete based on website_id, customer_group_id[, qty]
                if (strtolower($row['value']) == ValidatorInterface::DELETE_MARKER) {
                    $qtyCondition = $row['qty'] == ValidatorInterface::DELETE_QTY_VALUE
                        ? []
                        : [$this->_connection->quoteInto('qty = ?', $row['qty'])];
                    $this->countItemsDeleted += $this->_connection->delete(
                        $tableName,
                        array_merge(
                            [
                                $this->_connection->quoteInto($linkField . ' = ?', $productId),
                                $this->_connection->quoteInto('website_id = ?', $row['website_id']),
                                $this->_connection->quoteInto('customer_group_id = ?', $row['customer_group_id']),
                                $this->_connection->quoteInto('all_groups = ?', $row['all_groups'])
                            ],
                            $qtyCondition
                        )
                    );
                } else {
                    if ($this->decimalProcessor->hasSelfUpdate($row['value'])) {
                        $this->countItemsUpdated += $this->_connection->update(
                            $tableName,
                            ['value' => $row['value']],
                            [
                                $this->_connection->quoteInto("$linkField = ?", $productId),
                                $this->_connection->quoteInto('website_id = ?', $row['website_id']),
                                $this->_connection->quoteInto('customer_group_id = ?', $row['customer_group_id']),
                                $this->_connection->quoteInto('qty = ?', $row['qty'])
                            ]
                        );
                    } else {
                        $updateData[] = $row;
                    }
                }
            }
        }

        if ($updateData) {
            $this->countItemsUpdated += $this->_connection->insertOnDuplicate($tableName, $updateData, ['value']);
        }

        return $this;
    }

    protected function deleteTierPrices(array $data)
    {
        if (empty($data)) {
            return $this;
        }

        $tableName  = $this->getProductEntityTierPriceTableName();
        $linkField  = $this->getProductEntityLinkField();
        $oldSkus    = $this->getOldSku();
        foreach ($data as $sku => $priceRows) {
            if (!isset($oldSkus[strtolower($sku)])) {
                continue;
            }

            $productId = $oldSkus[strtolower($sku)][$linkField];
            foreach ($priceRows as $row) {
                $this->countItemsDeleted += $this->_connection->delete(
                    $tableName,
                    [
                        $this->_connection->quoteInto($linkField . ' = ?', $productId),
                        $this->_connection->quoteInto('website_id = ?', $row['website_id']),
                        $this->_connection->quoteInto('customer_group_id = ?', $row['customer_group_id']),
                        $this->_connection->quoteInto('all_groups = ?', $row['all_groups']),
                    ]
                );
            }
        }

        return $this;
    }

    /**
     * Check if product exists for specified SKU
     *
     * @param string $sku
     * @return bool
     */
    private function isSkuExist($sku)
    {
        $sku = strtolower($sku);
        return isset($this->_oldSku[$sku]);
    }

    /**
     * Get existing product data for specified SKU
     *
     * @param string $sku
     * @return array
     */
    private function getExistingSku($sku)
    {
        return $this->_oldSku[strtolower($sku)];
    }

    /**
     * Existing products SKU getter.
     *
     * Returns array of existing products data with SKU as key. All SKU keys are in lowercase for avoiding creation of
     * new products with the same SKU in different letter cases.
     *
     * @return array
     */
    public function getOldSku()
    {
        return $this->_oldSku;
    }

    public function getBasePrice($sku, $websiteId)
    {
        $sku = strtolower($sku);
        if (isset($this->_basePrices[$sku])) {
            if (isset($this->_basePrices[$sku][$websiteId])) {
                return $this->_basePrices[$sku][$websiteId];
            } elseif (isset($this->_basePrices[$sku][0])) {
                return $this->_basePrices[$sku][0];
            }
        }

        return null;
    }

    /**
     * Get product entity link field
     *
     * @return string
     * @throws \Exception
     */
    private function getProductEntityLinkField()
    {
        if (!$this->productEntityLinkField) {
            $this->productEntityLinkField = $this->getMetadataPool()
                ->getMetadata(ProductInterface::class)
                ->getLinkField();
        }
        return $this->productEntityLinkField;
    }

    private function getProductEntityTableName()
    {
        if (!$this->productEntityTableName) {
            $this->productEntityTableName = $this->_resourceFactory->create()->getTable(self::TABLE_MAIN);
        }
        return $this->productEntityTableName;
    }

    private function getProductEntityDecimalTableName()
    {
        if (!$this->productEntityDecimalTableName) {
            $this->productEntityDecimalTableName = $this->_resourceFactory->create()->getTable(self::TABLE_DECIMAL_PRICE);
        }
        return $this->productEntityDecimalTableName;
    }

    private function getProductEntityDatetimeTableName()
    {
        if (!$this->productEntityDatetimeTableName) {
            $this->productEntityDatetimeTableName = $this->_resourceFactory->create()->getTable(self::TABLE_DATETIME);
        }
        return $this->productEntityDatetimeTableName;
    }

    private function getProductEntityTierPriceTableName()
    {
        if (!$this->productEntityTierPriceTableName) {
            $this->productEntityTierPriceTableName = $this->_resourceFactory->create()->getTable(self::TABLE_TIER_PRICE);
        }
        return $this->productEntityTierPriceTableName;
    }

    /**
     * Set updated_at for product
     *
     * @param array $listSku
     * @return $this
     */
    protected function setUpdatedAt(array $listSku)
    {
        $updatedAt = $this->dateTimeDateTime->gmtDate('Y-m-d H:i:s');
        $this->_connection->update(
            $this->getProductEntityTableName(),
            [ProductInterface::UPDATED_AT => $updatedAt],
            $this->_connection->quoteInto('sku IN (?)', array_unique($listSku))
        );
        return $this;
    }

    /**
     * Get website id by code
     *
     * @param string $websiteCode
     * @return array|int|string
     */
    protected function getWebSiteId($websiteCode)
    {
        return $this->_catalogData->isPriceGlobal()
            ? 0
            : $this->_storeResolver->getWebsiteCodeToId($websiteCode);
    }

    /**
     * Get customer group id
     *
     * @param string $customerGroup
     * @return int
     */
    protected function getCustomerGroupId($customerGroup)
    {
        $customerGroups = $this->_getValidator(self::VALIDATOR_TIER_PRICE)->getCustomerGroups();
        $customerGroupMapping = array_flip($customerGroups);
        return $customerGroup == self::GROUP_ALL ? 0 : $customerGroupMapping[$customerGroup];
    }

    /**
     * Validator object getter.
     *
     * @param string $type
     * @return \MagePsycho\MipPricing\Model\Import\Pricing\Validator
     */
    protected function _getValidator($type)
    {
        return $this->_validators[$type];
    }

    /**
     * Add row as skipped
     *
     * @param int $rowNum
     * @param string $errorCode Error code or simply column name
     * @param string $errorLevel error level
     * @param string|null $colName optional column name
     * @return $this
     */
    private function skipRow(
        $rowNum,
        string $errorCode,
        string $errorLevel = ProcessingError::ERROR_LEVEL_NOT_CRITICAL,
        $colName = null
    ) {
        $this->addRowError($errorCode, $rowNum, $colName, null, $errorLevel);
        $this->getErrorAggregator()
            ->addRowToSkip($rowNum);
        return $this;
    }

    /**
     * Returns errorLevel for validation
     *
     * @param string $sku
     * @return string
     */
    private function getValidationErrorLevel($sku)
    {
        return (!$this->isSkuExist($sku) && Import::BEHAVIOR_REPLACE !== $this->getBehavior())
            ? ProcessingError::ERROR_LEVEL_CRITICAL
            : ProcessingError::ERROR_LEVEL_NOT_CRITICAL;
    }

    /**
     * Retrieve attribute by code
     *
     * @param string $attrCode
     * @return mixed
     */
    public function retrieveAttributeByCode($attrCode)
    {
        $attrCode = mb_strtolower($attrCode);

        if (!isset($this->_attributeCache[$attrCode])) {
            $this->_attributeCache[$attrCode] = $this->getResource()->getAttribute($attrCode);
        }

        return $this->_attributeCache[$attrCode];
    }

    /**
     * @return Proxy\Product\ResourceModel
     */
    protected function getResource()
    {
        if (!$this->_productResource) {
            $this->_productResource = $this->_resourceFactory->create();
        }
        return $this->_productResource;
    }
}
