<?php

namespace MagePsycho\MipPricing\Model\Import\Pricing;

use MagePsycho\MipPricing\Helper\Data as MipPricingHelper;
use MagePsycho\MipPricing\Model\System\Config\Source\RoundingType;
use MagePsycho\MipPricing\Model\Import\Pricing\Validator\AbstractPrice;

use MagePsycho\MassImporterPro\Model\Import\ErrorProcessing\ProcessingErrorAggregatorInterface;
use Magento\Catalog\Api\Data\ProductInterface;
use MagePsycho\MipPricing\Model\Import\Pricing;
use MagePsycho\MipPricing\Model\Import\Proxy\Product\ResourceModelFactory;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\EntityManager\MetadataPool;

/**
 * @category   MagePsycho
 * @package    MagePsycho_MipPricing
 * @author     Raj KB <magepsycho@gmail.com>
 * @website    https://www.magepsycho.com
 * @license    http://opensource.org/licenses/osl-3.0.php  Open Software License (OSL 3.0)
 */
class DecimalProcessor
{
    /**
     * @var MipPricingHelper
     */
    private $mipPricingHelper;

    /**
     * @var array
     */
    private $basePrices;

    /**
     * @var SkuProcessor
     */
    private $skuProcessor;

    /**
     * @var MetadataPool
     */
    private $metadataPool;

    /**
     * DB connection.
     *
     * @var \Magento\Framework\DB\Adapter\AdapterInterface
     */
    private $connection;

    /**
     * @var ResourceModelFactory
     */
    private $resourceFactory;

    /**
     * @var \Magento\CatalogImportExport\Model\Import\Proxy\Product\ResourceModel
     */
    private $resourceModel;

    /**
     * @var ProcessingErrorAggregatorInterface
     */
    private $errorAggregator;

    private $relativePriceAttributes = [
        Pricing::COL_SPECIAL_PRICE,
        Pricing::TIER_PRICE_PREFIX,
        Pricing::GROUP_PRICE_PREFIX
    ];

    /**
     * @var string
     */
    private $productEntityLinkField;

    private $productEntityTableName;
    private $productEntityDecimalTableName;
    private $storeTableName;
    private $basePriceAttributeId;

    public function __construct(
        MipPricingHelper $mipPricingHelper,
        SkuProcessor $skuProcessor,
        MetadataPool $metadataPool,
        ResourceConnection $resourceConnection,
        ResourceModelFactory $resourceModelFactory,
        ProcessingErrorAggregatorInterface $errorAggregator
    ) {
        $this->mipPricingHelper = $mipPricingHelper;
        $this->skuProcessor = $skuProcessor;
        $this->metadataPool = $metadataPool;
        $this->connection = $resourceConnection->getConnection();
        $this->resourceFactory = $resourceModelFactory;
        $this->errorAggregator = $errorAggregator;

        $this->initPriceResources();
    }

    private function initPriceResources()
    {
        if (null == $this->productEntityTableName) {
            $this->productEntityTableName = $this->getResource()->getTable('catalog_product_entity');
            $this->productEntityDecimalTableName = $this->getResource()->getTable('catalog_product_entity_decimal');
            $this->storeTableName = $this->getResource()->getTable('store');
            $relativePriceAttribute = $this->mipPricingHelper->getConfig()->getRelativePriceAttribute();
            $this->basePriceAttributeId = $this->getResource()->getAttribute($relativePriceAttribute)->getId();
        }
    }

    public function getBasePrices()
    {
        if (!$this->basePrices) {
            $this->basePrices = $this->_getBasePrices();
        }
        return $this->basePrices;
    }

    private function _getBasePrices()
    {
        $result = [];
        $select = $this->connection->select()->from(
            ['main' => $this->productEntityTableName],
            [
                'sku' => 'main.sku',
                $this->getProductEntityLinkField() => sprintf('main.%s', $this->getProductEntityLinkField())
            ]
        )->joinInner(
            ['price' => $this->productEntityDecimalTableName],
            sprintf(
                'main.%s = price.%s AND price.attribute_id = %s',
                $this->getProductEntityLinkField(),
                $this->getProductEntityLinkField(),
                $this->basePriceAttributeId
            ),
            ['store_id' => 'price.store_id', 'value' => 'price.value']
        )->joinLeft(
            ['store' => $this->storeTableName],
            'store.store_id = price.store_id',
            ['website_id', 'store.website_id']
        );

        foreach ($this->connection->fetchAll($select) as $price) {
            $result[strtolower($price['sku'])][$price['website_id']] = $price['value'];
        }
        return $result;
    }

    public function parseValue($value, $attrCode, $baseValue)
    {
        $value = $this->prepareExpression($value, $attrCode, $baseValue);
        $value = $this->prepareRounding($value);
        return $value;
    }

    protected function prepareExpression($value, $attrCode, $baseValue)
    {
        $sign    = $this->getSign($value);
        $percent = $this->isPercentageType($value);

        // prepare expression for self-increment/decrement (fixed/percent)
        if (in_array($sign, $this->getOperatorSigns())) {
            $value = substr($value, 1);
            if ($percent) {
                $value = substr($value, 0, -1);
            }
            $value = floatval($value);

            $value = $percent ? '`value` * ' . $value . '/100' : $value;
            $value = '`value`' . $sign . $value;
        }

        // Non-increment/decrement pricing (fixed(pure)/percent)
        if (in_array($attrCode, $this->relativePriceAttributes)
            && $percent && !in_array($sign, $this->getOperatorSigns())
        ) {
            if ($baseValue > 0) {
                $pc    = substr($value, 0, -1);
                $value = ($pc / 100) * $baseValue;
            }
        }
        return $value;
    }

    protected function prepareRounding($value, $options = [])
    {
        $roundingType = $options['price_rounding']
            ?? $this->mipPricingHelper->getConfig()->getRoundingType();

        if ($roundingType == RoundingType::ROUNDING_TYPE_NORMAL) {
            $value = 'ROUND(' . $value . ')';
        } elseif ($roundingType == RoundingType::ROUNDING_TYPE_TO_NEAREST) {
            $roundingNearest = $options['rounding_nearest']
                ?? $this->mipPricingHelper->getConfig()->getRoundingToNearestValue();
            $value           = 'FLOOR(' . $value . ') + ' . (float)$roundingNearest;
        }
        return $value;
    }

    protected function isRoundingEnabled($options = [])
    {
        $roundingType = $options['price_rounding']
            ?? $this->mipPricingHelper->getConfig()->getRoundingType();
        return !empty($roundingType)
            && in_array($roundingType, [RoundingType::ROUNDING_TYPE_NORMAL, RoundingType::ROUNDING_TYPE_TO_NEAREST]);
    }

    public function getSign($value)
    {
        return substr($value, 0, 1);
    }

    public function getOperatorSigns()
    {
        return [AbstractPrice::INCREMENT_OPERATOR, AbstractPrice::DECREMENT_OPERATOR];
    }

    public function isFixedType($value)
    {
        return strpos($value, '%') === false;
    }

    public function isPureNumeric($value)
    {
        return !$this->isRoundingEnabled() && $this->assertPureNumeric($value);
    }

    public function isPercentageType($value)
    {
        return strpos($value, '%') !== false;
    }

    /**
     * Check if value is pure numeric (integer or float) without any signs
     *
     * @param $value
     * @return false|int
     */
    public function assertPureNumeric($value)
    {
        return preg_match('/^(\d+\.)?(\d+)$/', $value);
    }

    /**
     * Check if value has self updating value:
     * `value` {*+-/} N
     * Self-update data can be only updated via UPDATE SQL statement
     *
     * @param $value
     * @return bool
     */
    public function hasSelfUpdate($value)
    {
        return strpbrk($value, '+-*/') !== false && strpos($value, 'value') !== false;
    }

    /**
     * @return string
     * @throws \Exception
     */
    private function getProductEntityLinkField()
    {
        if (!$this->productEntityLinkField) {
            $this->productEntityLinkField = $this->metadataPool->getMetadata(ProductInterface::class)->getLinkField();
        }

        return $this->productEntityLinkField;
    }

    /**
     * @return \Magento\CatalogImportExport\Model\Import\Proxy\Product\ResourceModel
     */
    private function getResource()
    {
        if (!$this->resourceModel) {
            $this->resourceModel = $this->resourceFactory->create();
        }

        return $this->resourceModel;
    }
}
