<?php

namespace UdeyTech\CategoryHelper\Helper;

use Magento\Catalog\Model\CategoryFactory;
use Magento\Catalog\Helper\Image;
use Magento\Framework\App\Helper\AbstractHelper;
use Magento\Store\Model\StoreManagerInterface;
use Magento\Catalog\Model\ResourceModel\Product\CollectionFactory;
use Magento\Catalog\Model\ResourceModel\Category\CollectionFactory as CategoryCollectionFactory;
use Magento\Framework\Registry;
use Magento\Framework\UrlInterface;
use Magento\Catalog\Api\ProductRepositoryInterface;

class Data extends AbstractHelper
{
    protected $categoryFactory;
    protected $storeManager;
    protected $imageHelper;
    protected $productCollectionFactory;
    protected $categoryCollectionFactory;
    protected $registry;
    protected $urlBuilder;
    protected $productRepository;

    public function __construct(
        CategoryFactory $categoryFactory,
        StoreManagerInterface $storeManager,
        Image $imageHelper,
        CollectionFactory $productCollectionFactory,
        CategoryCollectionFactory $categoryCollectionFactory,
        Registry $registry,
        UrlInterface $urlBuilder,
        ProductRepositoryInterface $productRepository
    ) {
        $this->categoryFactory = $categoryFactory;
        $this->storeManager = $storeManager;
        $this->imageHelper = $imageHelper;
        $this->productCollectionFactory = $productCollectionFactory;
        $this->categoryCollectionFactory = $categoryCollectionFactory;
        $this->registry = $registry;
        $this->urlBuilder = $urlBuilder;
        $this->productRepository = $productRepository;
    }

    /**
     * Get all first-level categories, their images, and listing page URLs
     *
     * @return array
     */
    public function getFirstLevelCategories()
    {
        $categories = [];
        $store = $this->storeManager->getStore();
        $rootCategoryId = $store->getRootCategoryId();
        $category = $this->categoryFactory->create()->load($rootCategoryId);

        $firstLevelCategories = $category->getChildrenCategories();

        foreach ($firstLevelCategories as $cat) {
            if ($cat->getIsActive()) {
                $category = $this->categoryFactory->create()->load($cat->getId());
                $imgUrl = $category->getImageUrl(); 
                
                if ($imgUrl) {
                    $categories[] = [
                        'name' => $cat->getName(),
                        'url' => $cat->getUrl(),
                        'image_url' => $imgUrl
                    ];
                } else {
                    $categories[] = [
                        'name' => $cat->getName(),
                        'url' => $cat->getUrl(),
                        'image_url' => null
                    ];
                }
            }
        }

        return $categories;
    }

        /**
     * Get a maximum of 6 third-level categories (without filtering by parent category ID)
     *
     * @return array
     */
    public function getThirdLevelCategories()
    {
        $categories = [];

        // Get all categories that are at the third level (level 3), limit to 6
        $categoryCollection = $this->categoryCollectionFactory->create();
        $categoryCollection->addAttributeToSelect('*')
            ->addIsActiveFilter()
            ->addFieldToFilter('level', 3) // Filter by category level (level 3 for third-level categories)
            ->setPageSize(8); // Limit the collection to 6 items

        // Loop through third-level categories
        foreach ($categoryCollection as $thirdLevelCategory) {
            $thirdLevelCategory = $this->categoryFactory->create()->load($thirdLevelCategory->getId());
            $imgUrl = $thirdLevelCategory->getImageUrl();

            // Prepare category details
            if ($imgUrl) {
                $categories[] = [
                    'name' => $thirdLevelCategory->getName(),
                    'url' => $thirdLevelCategory->getUrl(),
                    'image_url' => $imgUrl
                ];
            } else {
                $categories[] = [
                    'name' => $thirdLevelCategory->getName(),
                    'url' => $thirdLevelCategory->getUrl(),
                    'image_url' => null
                ];
            }
        }

        return $categories;
    }

    /**
     * Get products with the "Kitchen Sinks" attribute enabled
     *
     * @return \Magento\Catalog\Model\ResourceModel\Product\Collection
     */
    public function getKitchenSinksProducts()
    {
        return $this->getProductsByAttribute('kitchen_sinks');
    }

    /**
     * Get products with the "Bestsellers" attribute enabled
     *
     * @return \Magento\Catalog\Model\ResourceModel\Product\Collection
     */
    public function getBestsellersProducts()
    {
        return $this->getProductsByAttribute('bestsellers');
    }

    /**
     * Get products with the "Bath Enclosures" attribute enabled
     *
     * @return \Magento\Catalog\Model\ResourceModel\Product\Collection
     */
    public function getBathEnclosuresProducts()
    {
        return $this->getProductsByAttribute('bath_enclosures');
    }

    /**
     * Get products with the "Shower Baths" attribute enabled
     *
     * @return \Magento\Catalog\Model\ResourceModel\Product\Collection
     */
    public function getShowerBathsProducts()
    {
        return $this->getProductsByAttribute('shower_baths');
    }

    /**
     * Get products with the "Shower Baths" attribute enabled
     *
     * @return \Magento\Catalog\Model\ResourceModel\Product\Collection
     */
    public function getWhirlpoolBathsProducts()
    {
        return $this->getProductsByAttribute('whirlpool_baths');
    }

    /**
     * Get products with the "top_rated" attribute enabled
     *
     * @return \Magento\Catalog\Model\ResourceModel\Product\Collection
     */
    public function getTopRatedProducts()
    {
        return $this->getProductsByAttribute('top_rated');
    }

    /**
     * Get products by custom attribute
     *
     * @param string $attributeCode
     * @return \Magento\Catalog\Model\ResourceModel\Product\Collection
     */
    protected function getProductsByAttribute($attributeCode)
    {
        $collection = $this->productCollectionFactory->create();
        $collection->addAttributeToSelect('*');
        $collection->addAttributeToFilter($attributeCode, 1);
        $collection->setPageSize(6); // Limit the number of products
    
        // Load only visible products (excludes child products)
        $collection->addAttributeToFilter('visibility', ['neq' => \Magento\Catalog\Model\Product\Visibility::VISIBILITY_NOT_VISIBLE]);
    
        return $collection;
    }

    /**
     * Get child categories by parent category ID
     *
     * @param int $parentId
     * @return array
     */
    public function getChildCategories()
    {
        $categories = [];
        $categoryCollection = $this->categoryCollectionFactory->create();
        $categoryCollection->addAttributeToSelect('*')
            ->addIsActiveFilter()
            ->addAttributeToFilter('parent_id', $this->getCurrentCategoryId());

        foreach ($categoryCollection as $category) {
            $category = $this->categoryFactory->create()->load($category->getId());
            $imgUrl = $category->getImageUrl(); 
            
            if ($imgUrl) {
                $categories[] = [
                    'name' => $category->getName(),
                    'url' => $category->getUrl(),
                    'image_url' => $imgUrl
                ];
            } else {
                $categories[] = [
                    'name' => $category->getName(),
                    'url' => $category->getUrl(),
                    'image_url' => null
                ];
            }
        }

        return $categories;
    }

    /**
     * Get the current category ID from the registry
     *
     * @return int|null
     */
    public function getCurrentCategoryId()
    {
        $category = $this->registry->registry('current_category');
        return $category ? $category->getId() : null;
    }

    /**
     * Get Add to Cart URL for a product
     *
     * @param \Magento\Catalog\Model\Product $product
     * @return string
     */
    public function getAddToCartUrl($product)
    {
        return $this->urlBuilder->getUrl('checkout/cart/add', ['product' => $product->getId()]);
    }
}
