<?php

namespace MagePsycho\MipPricing\Console\Command;

use Magento\Framework\App\State;
use Magento\Framework\ObjectManagerInterface;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Filesystem\Exception\FileNotFoundException;
use Magento\Framework\Console\Cli;

/**
 * @category   MagePsycho
 * @package    MagePsycho_MipPricing
 * @author     Raj KB <magepsycho@gmail.com>
 * @website    https://www.magepsycho.com
 * @license    http://opensource.org/licenses/osl-3.0.php  Open Software License (OSL 3.0)
 */
class ImportCommand extends Command
{
    /**
     * @var ObjectManagerInterface
     */
    private $objectManager;

    /**
     * @var State
     */
    private $state;

    public function __construct(
        ObjectManagerInterface $objectManager,
        State $state
    ) {
        $this->objectManager = $objectManager;
        $this->state = $state;
        parent::__construct();
    }

    /**
     * {@inheritdoc}
     */
    protected function configure()
    {
        $this->setName('mip:pricing:import')
            ->setDescription('MIP Pricing Import via CLI')
            ->addArgument('filename', InputArgument::REQUIRED, 'CSV file path (relative or absolute)')
            ->addOption('behavior', 'b', InputOption::VALUE_OPTIONAL, 'Behavior (append|replace|delete)', 'append')
            ->addOption('reindex', 'ri', InputOption::VALUE_NONE, 'Reindex price after import')
            ->addOption('dry-validate', 'dv', InputOption::VALUE_NONE, 'Only validate data (no import)');
        parent::configure();
    }

    /**
     * {@inheritdoc}
     */
    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $this->state->setAreaCode('adminhtml');

        $import = $this->getImportModel();
        try {
            if ($input->getOption('behavior')) {
                $import->setBehavior($input->getOption('behavior'));
            }
            $import->setImportFile(realpath($input->getArgument('filename')));

            if ($input->getOption('dry-validate')) {
                $output->writeln($import->getFormattedLogTrace());
            } else {
                $result = $import->execute();
                if ($result) {
                    $output->writeln('<info>The pricing import was successful.</info>');
                    $output->writeln('Log trace:');
                    $output->writeln($import->getFormattedLogTrace());

                    // Reindex price
                    if ($input->getOption('reindex')) {
                        $output->writeln('<info>Reindexing catalog product price...</info>');
                        $import->reindex();
                        $output->writeln('<info>Completed.</info>');
                    }
                } else {
                    $output->writeln('<error>The pricing import failed.</error>');
                    $errors = $import->getErrors();
                    foreach ($errors as $error) {
                        $output->writeln(
                            '<error>' . $error->getErrorMessage() . ' - ' . $error->getErrorDescription() . '</error>'
                        );
                    }
                }
            }
            return Cli::RETURN_SUCCESS;
        } catch (FileNotFoundException $e) {
            $output->writeln('<error>File not found.</error>');
            return Cli::RETURN_FAILURE;
        } catch (\InvalidArgumentException $e) {
            $output->writeln('<error>Invalid source.</error>');
            $output->writeln('Log trace:');
            $output->writeln($import->getFormattedLogTrace());
            return Cli::RETURN_FAILURE;
        }
    }

    /**
     * @return \MagePsycho\MipPricing\Model\Console\Import
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    protected function getImportModel()
    {
        return $this->objectManager->create('MagePsycho\MipPricing\Model\Console\Import');
    }
}
