<?php

namespace MagePsycho\MipPricing\Model\Console;

use MagePsycho\MassImporterPro\Model\Import as MipImport;
use MagePsycho\MassImporterPro\Model\Import\ErrorProcessing\ProcessingErrorAggregatorInterface;
use MagePsycho\MipPricing\Model\Import\Pricing as PricingImport;
use MagePsycho\MassImporterPro\Model\System\Config\Source\ImportSource;
use MagePsycho\MassImporterPro\Model\System\Config\Source\ImportMethod;
use Symfony\Component\Filesystem\Exception\FileNotFoundException;

/**
 * @category   MagePsycho
 * @package    MagePsycho_MipPricing
 * @author     Raj KB <magepsycho@gmail.com>
 * @website    https://www.magepsycho.com
 * @license    http://opensource.org/licenses/osl-3.0.php  Open Software License (OSL 3.0)
 */
class Import
{
    const CATALOG_PRODUCT_PRICE_INDEXER = 'catalog_product_price';

    private $importFileInfo;

    /**
     * @var \MagePsycho\MassImporterPro\Model\Import
     */
    private $importModel;

    /**
     * @var \Magento\Framework\Filesystem\Directory\ReadFactory
     */
    private $readFactory;

    /**
     * @var \MagePsycho\MassImporterPro\Model\Import\Source\CsvFactory
     */
    private $csvSourceFactory;

    /**
     * @var \Magento\Indexer\Model\Indexer\CollectionFactory
     */
    private $indexerCollectionFactory;

    /**
     * @var \Magento\Indexer\Model\IndexerFactory
     */
    private $indexerFactory;

    public function __construct(
        \MagePsycho\MassImporterPro\Model\Import $importModel,
        \MagePsycho\MassImporterPro\Model\Import\Source\CsvFactory $csvSourceFactory,
        \Magento\Indexer\Model\Indexer\CollectionFactory $indexerCollectionFactory,
        \Magento\Framework\Filesystem\Directory\ReadFactory $readFactory,
        \Magento\Indexer\Model\IndexerFactory $indexerFactory
    ) {
        $this->importModel = $importModel;
        $this->csvSourceFactory = $csvSourceFactory;
        $this->indexerCollectionFactory = $indexerCollectionFactory;
        $this->readFactory = $readFactory;
        $this->indexerFactory = $indexerFactory;

        $this->importModel->setData(
            [
                'entity' => PricingImport::ENTITY_CODE,
                'behavior' => MipImport::BEHAVIOR_APPEND,
                'import_source' => ImportSource::IMPORT_SOURCE_FILE,
                'imported_via' => ImportMethod::IMPORT_METHOD_CLI,
                MipImport::FIELD_NAME_IMG_FILE_DIR => 'pub/media/catalog/product',
                MipImport::FIELD_NAME_VALIDATION_STRATEGY => ProcessingErrorAggregatorInterface::VALIDATION_STRATEGY_SKIP_ERRORS
            ]
        );
    }

    /**
     * @param string $filePath The Relative or Absolute path to import file
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function setImportFile($filePath)
    {
        if (!file_exists($filePath)) {
            throw new FileNotFoundException();
        }

        $this->importFileInfo = $pathInfo = pathinfo($filePath);
        $validate = $this->importModel->validateSource(
            $this->csvSourceFactory->create(
                [
                    'file' => $pathInfo['basename'],
                    'directory' => $this->readFactory->create($pathInfo['dirname'])
                ]
            )
        );
        if (!$validate) {
            throw new \InvalidArgumentException();
        }
    }

    private function getImportFileInfo()
    {
        return $this->importFileInfo;
    }

    /**
     * @param $behavior
     */
    public function setBehavior($behavior)
    {
        if (in_array(
                $behavior,
                [
                    MipImport::BEHAVIOR_APPEND,
                    MipImport::BEHAVIOR_ADD_UPDATE,
                    MipImport::BEHAVIOR_REPLACE,
                    MipImport::BEHAVIOR_DELETE
                ]
            )
        ) {
            $this->importModel->setData('behavior', $behavior);
        }
    }

    /**
     * @return bool
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function execute()
    {
        $this->importModel->createCliHistoryReport($this->getImportFileInfo());

        $result = $this->importModel->importSource();
        if ($result) {
            $this->importModel->invalidateIndex();
        }
        return $result;
    }

    /**
     * @return string
     */
    public function getFormattedLogTrace()
    {
        return $this->importModel->getFormatedLogTrace();
    }

    /**
     * @return MipImport\ErrorProcessing\ProcessingError[]
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function getErrors()
    {
        return $this->importModel->getErrorAggregator()->getAllErrors();
    }

    public function reindex($indexId = self::CATALOG_PRODUCT_PRICE_INDEXER)
    {
        $indexer = $this->indexerFactory->create()->load($indexId);
        $indexer->reindexAll();
    }
}
