<?php

namespace MagePsycho\MipPricing\Model\Export;

use MagePsycho\MipPricing\Model\Import\Pricing as ImportPricing;
use MagePsycho\MipPricing\Model\Import\Pricing\Validator\TierPrice as TierPriceValidator;
use Magento\Catalog\Model\Product as CatalogProduct;
use Magento\Store\Model\Store;

/**
 * @category   MagePsycho
 * @package    MagePsycho_MipPricing
 * @author     Raj KB <magepsycho@gmail.com>
 * @website    https://www.magepsycho.com
 * @license    http://opensource.org/licenses/osl-3.0.php  Open Software License (OSL 3.0)
 */
class Pricing extends \Magento\CatalogImportExport\Model\Export\Product
{
    const ENTITY_MIP_PRICING = 'mip_pricing';

    /**
     * @var \Magento\CatalogImportExport\Model\Import\Product\StoreResolver
     */
    protected $storeResolver;

    /**
     * @var \Magento\Customer\Api\GroupRepositoryInterface
     */
    protected $groupRepository;

    /**
     * @var \Magento\Framework\App\ResourceConnection
     */
    protected $resource;

    /**
     * @var string
     */
    protected $_entityTypeCode;

    /**
     * Customer groups ID-to-name.
     *
     * @var array
     */
    protected $customerGroups = [];

    /**
     * @var array
     */
    protected $headerTemplates = [];

    /**
     * @var int
     */
    protected $_storeId;

    /**
     * @var \Magento\Framework\Api\SearchCriteriaBuilder
     */
    private $searchCriteriaBuilder;

    public function __construct(
        \Magento\Framework\Stdlib\DateTime\TimezoneInterface $localeDate,
        \Magento\Eav\Model\Config $config,
        \Magento\Framework\App\ResourceConnection $resource,
        \Magento\Store\Model\StoreManagerInterface $storeManager,
        \Psr\Log\LoggerInterface $logger,
        \Magento\Catalog\Model\ResourceModel\Product\CollectionFactory $collectionFactory,
        \Magento\ImportExport\Model\Export\ConfigInterface $exportConfig,
        \Magento\Catalog\Model\ResourceModel\ProductFactory $productFactory,
        \Magento\Eav\Model\ResourceModel\Entity\Attribute\Set\CollectionFactory $attrSetColFactory,
        \Magento\Catalog\Model\ResourceModel\Category\CollectionFactory $categoryColFactory,
        \Magento\CatalogInventory\Model\ResourceModel\Stock\ItemFactory $itemFactory,
        \Magento\Catalog\Model\ResourceModel\Product\Option\CollectionFactory $optionColFactory,
        \Magento\Catalog\Model\ResourceModel\Product\Attribute\CollectionFactory $attributeColFactory,
        \Magento\CatalogImportExport\Model\Export\Product\Type\Factory $_typeFactory,
        \Magento\Catalog\Model\Product\LinkTypeProvider $linkTypeProvider,
        \Magento\CatalogImportExport\Model\Export\RowCustomizerInterface $rowCustomizer,
        \Magento\CatalogImportExport\Model\Import\Product\StoreResolver $storeResolver,
        \Magento\Customer\Api\GroupRepositoryInterface $groupRepository,
        \Magento\Framework\Api\SearchCriteriaBuilder $searchCriteriaBuilder
    ) {
        $this->storeResolver            = $storeResolver;
        $this->groupRepository          = $groupRepository;
        $this->resource                 = $resource;
        $this->searchCriteriaBuilder    = $searchCriteriaBuilder;

        parent::__construct(
            $localeDate,
            $config,
            $resource,
            $storeManager,
            $logger,
            $collectionFactory,
            $exportConfig,
            $productFactory,
            $attrSetColFactory,
            $categoryColFactory,
            $itemFactory,
            $optionColFactory,
            $attributeColFactory,
            $_typeFactory,
            $linkTypeProvider,
            $rowCustomizer
        );

        $this->initCustomerGroups();
    }

    /**
     * @return $this
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    protected function initTypeModels()
    {
        $productTypes = $this->_exportConfig->getEntityTypes(CatalogProduct::ENTITY);
        foreach ($productTypes as $productTypeName => $productTypeConfig) {
            if (!($model = $this->_typeFactory->create($productTypeConfig['model']))) {
                throw new \Magento\Framework\Exception\LocalizedException(
                    __('Entity type model \'%1\' is not found', $productTypeConfig['model'])
                );
            }
            if (!$model instanceof \Magento\CatalogImportExport\Model\Export\Product\Type\AbstractType) {
                throw new \Magento\Framework\Exception\LocalizedException(
                    __(
                        'Entity type model must be an instance of'
                        . ' \Magento\CatalogImportExport\Model\Export\Product\Type\AbstractType'
                    )
                );
            }
            if ($model->isSuitable()) {
                $this->_productTypeModels[$productTypeName] = $model;
                $this->_disabledAttrs = array_merge($this->_disabledAttrs, $model->getDisabledAttrs());
                $this->_indexValueAttributes = array_merge(
                    $this->_indexValueAttributes,
                    $model->getIndexValueAttributes()
                );
            }
        }
        if (!$this->_productTypeModels) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __('There are no product types available for export')
            );
        }
        $this->_disabledAttrs = array_unique($this->_disabledAttrs);
        return $this;
    }

    protected function initHeaderTemplates()
    {
        $permanentHeaderTemplates = [
            ImportPricing::COL_SKU                   => '',
            ImportPricing::COL_WEBSITE_ID            => '',
        ];

        $mipExportFields = $this->getMipExportFields();

        $exportTierPriceFlag    = false;
        $exportHeaderTemplates  = [];
        if (!empty($mipExportFields)) {
            $mipExportFields = explode(',', $mipExportFields);
            foreach ($mipExportFields as $mipExportField) {
                if ($mipExportField == 'name') {
                    $exportHeaderTemplates[ImportPricing::COL_NAME] = '';
                } elseif ($mipExportField == 'cost') {
                    $exportHeaderTemplates[ImportPricing::COL_COST] = '';
                } elseif ($mipExportField == 'price') {
                    $exportHeaderTemplates[ImportPricing::COL_PRICE] = '';
                } elseif ($mipExportField == 'msrp') {
                    $exportHeaderTemplates[ImportPricing::COL_MSRP] = '';
                } elseif ($mipExportField == 'special_price') {
                    $exportHeaderTemplates[ImportPricing::COL_SPECIAL_PRICE]        = '';
                    $exportHeaderTemplates[ImportPricing::COL_SPECIAL_FROM_DATE]    = '';
                    $exportHeaderTemplates[ImportPricing::COL_SPECIAL_TO_DATE]      = '';
                } elseif ($mipExportField == 'tier_price') {
                    $exportTierPriceFlag = true;
                }
            }
        } else {
            $exportHeaderTemplates = [
                ImportPricing::COL_NAME                  => '',
                ImportPricing::COL_COST                  => '',
                ImportPricing::COL_PRICE                 => '',
                ImportPricing::COL_MSRP                  => '',
                ImportPricing::COL_SPECIAL_PRICE         => '',
                ImportPricing::COL_SPECIAL_FROM_DATE     => '',
                ImportPricing::COL_SPECIAL_TO_DATE       => '',
            ];
        }

        $this->headerTemplates = array_merge($permanentHeaderTemplates, $exportHeaderTemplates);

        if (empty($mipExportFields) || $exportTierPriceFlag) {
            $customerGroups = array_merge([0 => ImportPricing::GROUP_ALL], $this->getCustomerGroups());
            foreach ($customerGroups as $groupId => $groupName) {
                $this->headerTemplates = array_merge($this->headerTemplates, ['tier_price:' . $groupName => '']);
            }
        }
    }

    protected function getMipExportFields()
    {
        $parameters   = isset($this->_parameters['mip_pricing_attributes'])
            ? $this->_parameters
            : $this->_parameters[\Magento\ImportExport\Model\Export::FILTER_ELEMENT_GROUP];
        return $parameters['mip_pricing_attributes'] ?? [];
    }

    protected function getWebsiteId()
    {
        $parameters = isset($this->_parameters['mip_website_id'])
            ? $this->_parameters
            : $this->_parameters[\Magento\ImportExport\Model\Export::FILTER_ELEMENT_GROUP];
        return $parameters['mip_website_id'] ?? [];
    }

    protected function getStoreId()
    {
        if ($this->_storeId === null) {
            $websiteId = $this->getWebsiteId();
            $this->_storeId = $websiteId > 0
                ? $this->_storeManager->getWebsite($websiteId)->getDefaultStore()->getId()
                : Store::DEFAULT_STORE_ID;
        }
        return $this->_storeId;
    }

    protected function initCustomerGroups()
    {
        foreach ($this->groupRepository->getList($this->searchCriteriaBuilder->create())->getItems() as $group) {
            $this->customerGroups[$group->getId()] = $group->getCode();
        }
    }

    /**
     * Export process
     *
     * @return string
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function export()
    {
        $this->initHeaderTemplates();

        //Execution time may be very long
        set_time_limit(0);

        $writer = $this->getWriter();
        $page = 0;
        while (true) {
            ++$page;
            $entityCollection = $this->_getEntityCollection(true);
            $entityCollection->setOrder('has_options', 'asc');
            $entityCollection->setStoreId(
                $this->getStoreId()
            );
            $this->_prepareEntityCollection($entityCollection);
            $this->paginateCollection($page, $this->getItemsPerPage());
            if ($entityCollection->count() == 0) {
                break;
            }
            $exportData = $this->getExportData();
            foreach ($exportData as $dataRow) {
                $writer->writeRow($dataRow);
            }
            if ($entityCollection->getCurPage() >= $entityCollection->getLastPageNumber()) {
                break;
            }
        }
        return $writer->getContents();
    }

    protected function getExportData()
    {
        $exportData = [];
        try {
            $products = $this->loadStoreCollection($this->getStoreId());
            if (!empty($products)) {
                $exportData = $this->prepareExportData($products);
                if (!empty($exportData)) {
                    asort($exportData);
                }
            }
        } catch (\Throwable $e) {
            $this->_logger->critical($e);
        }

        return $exportData;
    }

    protected function loadStoreCollection($storeId)
    {
        $data = [];
        $collection = $this->_getEntityCollection();
        $collection->setOrder('entity_id', 'asc');
        $this->_prepareEntityCollection($collection);
        $collection->setStoreId($storeId);
        $collection->load();
        foreach ($collection as $itemId => $item) {
            $data[$itemId][$storeId] = $item;
        }
        $collection->clear();
        return $data;
    }

    /**
     * Prepare data for export.
     *
     * @param array $productsData Products to export.
     * @return array Export rows to display.
     */
    private function prepareExportData(array $productsData)
    {
        $exportData = [];
        foreach ($productsData as $productId => $productData) {
            $product = $productData[$this->getStoreId()];
            $priceCols = [
                ImportPricing::COL_SKU                   => $product->getSku(),
                ImportPricing::COL_NAME                  => $product->getName(),
                ImportPricing::COL_WEBSITE_ID            => $this->getWebsiteId(),
                ImportPricing::COL_COST                  => $this->formatPrice($product->getData('cost')),
                ImportPricing::COL_PRICE                 => $this->formatPrice($product->getData('price')),
                ImportPricing::COL_MSRP                  => $this->formatPrice($product->getData('msrp')),
                ImportPricing::COL_SPECIAL_PRICE         => $this->formatPrice($product->getData('special_price')),
                ImportPricing::COL_SPECIAL_FROM_DATE     => $product->getData('special_from_date'),
                ImportPricing::COL_SPECIAL_TO_DATE       => $product->getData('special_to_date')
            ];

            $tierPriceCols = [];
            $product->getTierPrice();
            $tierPrices = $product->getData('tier_price') ?? [];
            foreach ($tierPrices as $tierPrice) {
                $priceValue = !empty($tierPrice['percentage_value'])
                    ? ((int) $tierPrice['percentage_value']) . '%'
                    : $this->formatPrice($tierPrice['price']);
                $groupLabel = $this->getCustomerGroupLabel($tierPrice['cust_group']);
                $tierPriceCols['tier_price:' . $groupLabel][] = ((int) $tierPrice['price_qty'])
                    . TierPriceValidator::QTY_TO_PRICE_SEPARATOR . $priceValue;
            }
            $exportData[] = array_merge($priceCols, $tierPriceCols);
        }

        //Formatting data for export.
        $customExportData = [];
        foreach ($exportData as $row) {
            $customExportData[] = $this->createExportRow($row);
        }

        return $customExportData;
    }

    private function formatPrice($price)
    {
        if ($price > 0) {
            $price = sprintf("%.2f", $price);
        }
        return $price;
    }

    /**
     * Creating export-formatted row
     *
     * @param array $exportRow
     * @return array
     */
    private function createExportRow(array $exportRow)
    {
        $_exportRow = $this->headerTemplates;
        foreach (array_keys($_exportRow) as $keyCol) {
            if (array_key_exists($keyCol, $exportRow)) {
                $_exportRow[$keyCol] = is_array($exportRow[$keyCol])
                    ? implode(TierPriceValidator::MULTI_VALUES_SEPARATOR, $exportRow[$keyCol])
                    : $exportRow[$keyCol];
            }
        }
        return $_exportRow;
    }

    /**
     * Get existing customers groups
     *
     * @return array
     */
    public function getCustomerGroups()
    {
        if (!$this->customerGroups) {
            $this->initCustomerGroups();
        }
        return $this->customerGroups;
    }

    /**
     * Get customer group label
     *
     * @param int $customerGroupId
     * @return string
     */
    protected function getCustomerGroupLabel($customerGroupId)
    {
        // groupId 32000 represents 'All Groups'
        return $this->getCustomerGroups()[$customerGroupId] ?? ImportPricing::GROUP_ALL;
    }

    /**
     * Get Entity type code
     *
     * @return string
     */
    public function getEntityTypeCode()
    {
        if (!$this->_entityTypeCode) {
            $this->_entityTypeCode = CatalogProduct::ENTITY;
        } else {
            $this->_entityTypeCode = self::ENTITY_MIP_PRICING;
        }
        return $this->_entityTypeCode;
    }
}
