<?php
/**
 * Copyright © MageWorx. All rights reserved.
 * See LICENSE.txt for license details.
 */
declare(strict_types=1);

namespace MageWorx\OptionFeatures\Plugin;


use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\ObjectManagerInterface;
use Magento\Framework\Registry;
use MageWorx\OptionFeatures\Model\QtyMultiplier;
use Magento\InventorySalesApi\Api\Data\SalesEventInterface;

class QtyMultiplierSourceDeductionService
{
    /**
     * @var Registry
     */
    protected $registry;

    /**
     * @var QtyMultiplier
     */
    protected $qtyMultiplier;

    /**
     * Object Manager instance
     *
     * @var ObjectManagerInterface
     */
    protected $objectManager = null;

    /**
     * @var ProductRepositoryInterface
     */
    protected $productRepository;

    /**
     * QtyMuliptilerSourceDeductionService constructor.
     *
     * @param Registry $registry
     * @param QtyMultiplier $qtyMultiplier
     * @param ObjectManagerInterface $objectManager
     * @param ProductRepositoryInterface $productRepository
     */
    public function __construct(
        Registry $registry,
        QtyMultiplier $qtyMultiplier,
        ObjectManagerInterface $objectManager,
        ProductRepositoryInterface $productRepository
    ) {
        $this->registry          = $registry;
        $this->qtyMultiplier     = $qtyMultiplier;
        $this->objectManager     = $objectManager;
        $this->productRepository = $productRepository;
    }

    /**
     * @inheritdoc
     */
    public function beforeExecute(
        \Magento\InventorySourceDeductionApi\Model\SourceDeductionService $subject,
        \Magento\InventorySourceDeductionApi\Model\SourceDeductionRequestInterface $sourceDeductionRequest
    ): void {

        $salesEvent = $sourceDeductionRequest->getSalesEvent()->getType();

        if ($salesEvent == SalesEventInterface::EVENT_CREDITMEMO_CREATED) {
            $dataToDeduct = $this->registry->registry('current_creditmemo');
            $multiplier   = -1;
        } elseif ($salesEvent == SalesEventInterface::EVENT_SHIPMENT_CREATED) {
            $dataToDeduct = $this->registry->registry('current_shipment');
            $multiplier   = 1;
        } else {
            return;
        }

        $sourceItemsSave                 = $this->objectManager->get(
            \Magento\InventoryApi\Api\SourceItemsSaveInterface::class
        );
        $getStockItemConfiguration       = $this->objectManager->get(
            \Magento\InventoryConfigurationApi\Api\GetStockItemConfigurationInterface::class
        );
        $getStockBySalesChannel          = $this->objectManager->get(
            \Magento\InventorySalesApi\Api\GetStockBySalesChannelInterface::class
        );
        $getSourceItemBySourceCodeAndSku = $this->objectManager->get(
            \Magento\InventorySourceDeductionApi\Model\GetSourceItemBySourceCodeAndSku::class
        );

        $currentItems = $dataToDeduct->getAllItems();
        $sourceItems  = [];
        $sourceCode   = $sourceDeductionRequest->getSourceCode();
        $salesChannel = $sourceDeductionRequest->getSalesChannel();
        $stockId      = $getStockBySalesChannel->execute($salesChannel)->getStockId();
        $qtyToDeduct  = [];
        foreach ($currentItems as $item) {
            $product                = $this->productRepository->getById($item->getProductId());
            $itemSku                = $product->getSku();
            $qty                    = $item->getQty();
            $stockItemConfiguration = $getStockItemConfiguration->execute(
                $itemSku,
                $stockId
            );

            if (!$stockItemConfiguration->isManageStock()) {
                //We don't need to Manage Stock
                continue;
            }

            $currentQtyMultiplierQty = $this->qtyMultiplier->getQtyMultiplierQtyForCurrentItemQty(
                $item->getOrderItem(),
                $qty
            );

            if (!$currentQtyMultiplierQty) {
                continue;
            }

            $qty = ($currentQtyMultiplierQty - $qty) * ($multiplier);

            if (isset($qtyToDeduct[$itemSku])) {
                $qtyToDeduct[$itemSku] += $qty;
            } else {
                $qtyToDeduct[$itemSku] = $qty;
            }
        }

        foreach ($currentItems as $item) {
            $itemSku = $product->getSku();

            if (!array_key_exists($itemSku, $qtyToDeduct)) {
                continue;
            }
            $sourceItem = $getSourceItemBySourceCodeAndSku->execute($sourceCode, $itemSku);
            if (($sourceItem->getQuantity() - $qtyToDeduct[$itemSku]) >= 0) {
                $sourceItem->setQuantity($sourceItem->getQuantity() - $qtyToDeduct[$itemSku]);
                $sourceItems[] = $sourceItem;
            } else {
                throw new LocalizedException(
                    __('Not all of your products are available in the requested quantity.')
                );
            }
        }

        if (!empty($sourceItems)) {
            $sourceItemsSave->execute($sourceItems);
        }
    }
}
